@@ -1,5 +1,5 @@
  /* BoCA - BonkEnc Component Architecture
-  * Copyright (C) 2007-2020 Robert Kausch <robert.kausch@freac.org>
+  * Copyright (C) 2007-2021 Robert Kausch <robert.kausch@freac.org>
   *
   * This program is free software; you can redistribute it and/or
   * modify it under the terms of the GNU General Public License as
@@ -133,8 +133,8 @@ Error BoCA::DecoderCueSheet::GetStreamInfo(const String &streamURI, Track &track
 		{
 			String	 genre;
 
-			if (line.Contains("\"")) genre = line.SubString(line.Find("\"") + 1, line.FindLast("\"") - line.Find("\"") - 1);
-			else			 genre = line.Tail(line.Length() - 10);
+			if (line.Contains("\"")) genre = UnescapeString(line.SubString(line.Find("\"") + 1, line.FindLast("\"") - line.Find("\"") - 1));
+			else			 genre = UnescapeString(line.Tail(line.Length() - 10));
 
 			if (!readInfoTags || preferCueSheets) info.genre = genre;
 
@@ -147,8 +147,8 @@ Error BoCA::DecoderCueSheet::GetStreamInfo(const String &streamURI, Track &track
 		{
 			String	 comment;
 
-			if (line.Contains("\"")) comment = line.SubString(line.Find("\"") + 1, line.FindLast("\"") - line.Find("\"") - 1);
-			else			 comment = line.Tail(line.Length() - 12);
+			if (line.Contains("\"")) comment = UnescapeString(line.SubString(line.Find("\"") + 1, line.FindLast("\"") - line.Find("\"") - 1));
+			else			 comment = UnescapeString(line.Tail(line.Length() - 12));
 
 			if (!readInfoTags || preferCueSheets) info.comment = comment;
 
@@ -172,8 +172,8 @@ Error BoCA::DecoderCueSheet::GetStreamInfo(const String &streamURI, Track &track
 		{
 			String	 artist;
 
-			if (line.Contains("\"")) artist = line.SubString(line.Find("\"") + 1, line.FindLast("\"") - line.Find("\"") - 1);
-			else			 artist = line.Tail(line.Length() - 10);
+			if (line.Contains("\"")) artist = UnescapeString(line.SubString(line.Find("\"") + 1, line.FindLast("\"") - line.Find("\"") - 1));
+			else			 artist = UnescapeString(line.Tail(line.Length() - 10));
 
 			if (!readInfoTags || preferCueSheets)
 			{
@@ -190,36 +190,20 @@ Error BoCA::DecoderCueSheet::GetStreamInfo(const String &streamURI, Track &track
 		{
 			String	 songwriter;
 
-			if (line.Contains("\"")) songwriter = line.SubString(line.Find("\"") + 1, line.FindLast("\"") - line.Find("\"") - 1);
-			else			 songwriter = line.Tail(line.Length() - 11);
+			if (line.Contains("\"")) songwriter = UnescapeString(line.SubString(line.Find("\"") + 1, line.FindLast("\"") - line.Find("\"") - 1));
+			else			 songwriter = UnescapeString(line.Tail(line.Length() - 11));
 
-			if (!readInfoTags || preferCueSheets)
-			{
-				for (Int i = 0; i < info.other.Length(); i++)
-				{
-					if (info.other.GetNth(i).StartsWith(String(INFO_COMPOSER).Append(":"))) info.other.RemoveNth(i);
-				}
+			if (!readInfoTags || preferCueSheets) info.SetOtherInfo(INFO_COMPOSER, songwriter);
 
-				info.other.Add(String(INFO_COMPOSER).Append(":").Append(songwriter));
-			}
-
-			if (!trackMode && !dataMode)
-			{
-				for (Int i = 0; i < albumInfo.other.Length(); i++)
-				{
-					if (albumInfo.other.GetNth(i).StartsWith(String(INFO_COMPOSER).Append(":"))) albumInfo.other.RemoveNth(i);
-				}
-
-				albumInfo.other.Add(String(INFO_COMPOSER).Append(":").Append(songwriter));
-			}
+			if (!trackMode && !dataMode) albumInfo.SetOtherInfo(INFO_COMPOSER, songwriter);
 		}
 
 		if (line.StartsWith("TITLE "))
 		{
 			String	 title;
 
-			if (line.Contains("\"")) title = line.SubString(line.Find("\"") + 1, line.FindLast("\"") - line.Find("\"") - 1);
-			else			 title = line.Tail(line.Length() - 6);
+			if (line.Contains("\"")) title = UnescapeString(line.SubString(line.Find("\"") + 1, line.FindLast("\"") - line.Find("\"") - 1));
+			else			 title = UnescapeString(line.Tail(line.Length() - 6));
 
 			if (!readInfoTags || preferCueSheets)
 			{
@@ -239,6 +223,20 @@ Error BoCA::DecoderCueSheet::GetStreamInfo(const String &streamURI, Track &track
 			if (Info::IsISRC(isrc)) info.isrc = isrc;
 		}
 
+		/* Flags.
+		 */
+		if (line.StartsWith("FLAGS "))
+		{
+			if (!trackMode) continue;
+
+			const Array<String>	&flags = line.Tail(line.Length() - 6).Explode(" ");
+
+			foreach (const String &flag, flags)
+			{
+				if (flag == "PRE") info.SetOtherInfo(INFO_PRE_EMPHASIS, "1");
+			}
+		}
+
 		/* Track sample offset.
 		 */
 		if (line.StartsWith("INDEX 01 "))
@@ -287,7 +285,7 @@ Error BoCA::DecoderCueSheet::GetStreamInfo(const String &streamURI, Track &track
 										       msf.GetNth(1).ToInt()	  * format.rate +
 										       msf.GetNth(2).ToInt()	  * format.rate / 75, fileLength);
 
-							if (samplePos >= iTrack.sampleOffset)
+							if (samplePos > iTrack.sampleOffset)
 							{
 								iTrack.length	= samplePos - iTrack.sampleOffset;
 								iTrack.fileSize	= Math::Round(Float(iTrack.fileSize) / fileLength * iTrack.length);
@@ -303,7 +301,7 @@ Error BoCA::DecoderCueSheet::GetStreamInfo(const String &streamURI, Track &track
 
 			/* Add previous track.
 			 */
-			if (iTrack.length != -1 || iTrack.approxLength != -1)
+			if (iTrack.length > 0 || iTrack.approxLength > 0)
 			{
 				iTrack.SetFormat(format);
 				iTrack.SetInfo(info);
@@ -412,10 +410,13 @@ Error BoCA::DecoderCueSheet::GetStreamInfo(const String &streamURI, Track &track
 			 */
 			Config	*decoderConfig = Config::Copy(GetConfiguration());
 
-			decoderConfig->SetIntValue("Tags", "ReadChapters", False);
-			decoderConfig->SetIntValue("Tags", "ReadEmbeddedCueSheets", False);
+			if (trackMode)
+			{
+				decoderConfig->SetIntValue("Tags", "ReadChapters", False);
+				decoderConfig->SetIntValue("Tags", "ReadEmbeddedCueSheets", False);
 
-			decoder->SetConfiguration(decoderConfig);
+				decoder->SetConfiguration(decoderConfig);
+			}
 
 			/* Get stream info.
 			 */
@@ -581,18 +582,7 @@ Void BoCA::DecoderCueSheet::UpdateInfoWithAlbumInfo(Info &info, const Info &albu
 	if (albumInfo.genre   != NIL) info.genre   = albumInfo.genre;
 	if (albumInfo.comment != NIL) info.comment = albumInfo.comment;
 
-	for (Int i = 0; i < albumInfo.other.Length(); i++)
-	{
-		if (albumInfo.other.GetNth(i).StartsWith(String(INFO_COMPOSER).Append(":")))
-		{
-			for (Int j = 0; j < info.other.Length(); j++)
-			{
-				if (info.other.GetNth(j).StartsWith(String(INFO_COMPOSER).Append(":"))) info.other.RemoveNth(j);
-			}
-
-			info.other.Add(albumInfo.other.GetNth(i));
-		}
-	}
+	if (albumInfo.HasOtherInfo(INFO_COMPOSER)) info.SetOtherInfo(INFO_COMPOSER, albumInfo.GetOtherInfo(INFO_COMPOSER));
 
 	if (albumInfo.album_gain != NIL) info.album_gain = albumInfo.album_gain;
 	if (albumInfo.album_peak != NIL) info.album_peak = albumInfo.album_peak;
@@ -626,6 +616,11 @@ Bool BoCA::DecoderCueSheet::AddTrack(const Track &track, Array<Track> &tracks) c
 	return True;
 }
 
+String BoCA::DecoderCueSheet::UnescapeString(const String &string)
+{
+	return string.Replace(" // ", "\n").Replace("''", "\"");
+}
+
 BoCA::DecoderCueSheet::DecoderCueSheet()
 {
 	configLayer = NIL;
