@@ -1,5 +1,5 @@
  /* BoCA - BonkEnc Component Architecture
-  * Copyright (C) 2007-2020 Robert Kausch <robert.kausch@freac.org>
+  * Copyright (C) 2007-2021 Robert Kausch <robert.kausch@freac.org>
   *
   * This program is free software; you can redistribute it and/or
   * modify it under the terms of the GNU General Public License as
@@ -15,6 +15,10 @@
 
 #include "faad2.h"
 
+#ifndef PS
+#	define PS 29
+#endif
+
 using namespace smooth::IO;
 
 const String &BoCA::DecoderFAAD2::GetComponentSpecs()
@@ -45,6 +49,7 @@ const String &BoCA::DecoderFAAD2::GetComponentSpecs()
 			      <extension>m4r</extension>				\
 			      <extension>mp4</extension>				\
 			      <extension>3gp</extension>				\
+			      <extension>3gpp</extension>				\
 			      <tag id=\"mp4-tag\" mode=\"other\">MP4 Metadata</tag>	\
 			    </format>							\
 											\
@@ -91,7 +96,7 @@ Bool BoCA::DecoderFAAD2::CanOpenStream(const String &streamURI)
 
 	if (mp4v2dll != NIL && (in.InputNumberRaw(8) & 0xFFFFFFFF) == 'ftyp')
 	{
-		MP4FileHandle	 mp4File  = ex_MP4Read(streamURI.ConvertTo("UTF-8"));
+		MP4FileHandle	 mp4File  = ex_MP4ReadProvider(streamURI.ConvertTo("UTF-8"), NIL);
 		Int		 mp4Track = GetAudioTrack(mp4File);
 
 		if (mp4Track >= 0 && ex_MP4GetSampleSize(mp4File, mp4Track, 1) > 0)
@@ -100,10 +105,10 @@ Bool BoCA::DecoderFAAD2::CanOpenStream(const String &streamURI)
 
 			if (type == MAIN   ||
 			    type == LC	   ||
-			    type == SSR	   ||
 			    type == LTP	   ||
 			    type == HE_AAC ||
 			    type == LD	   ||
+			    type == PS	   ||
 
 			    type == ER_LC  ||
 			    type == ER_LTP) isValidFile = True;
@@ -131,7 +136,7 @@ Error BoCA::DecoderFAAD2::GetStreamInfo(const String &streamURI, Track &track)
 		track.fileSize	= File(streamURI).GetFileSize();
 		track.length	= -1;
 
-		MP4FileHandle	 mp4File  = ex_MP4Read(streamURI.ConvertTo("UTF-8"));
+		MP4FileHandle	 mp4File  = ex_MP4ReadProvider(streamURI.ConvertTo("UTF-8"), NIL);
 		Int		 mp4Track = GetAudioTrack(mp4File);
 
 		if (mp4Track >= 0 && ex_MP4GetSampleSize(mp4File, mp4Track, 1) > 0)
@@ -152,7 +157,11 @@ Error BoCA::DecoderFAAD2::GetStreamInfo(const String &streamURI, Track &track)
 
 			ex_MP4GetTrackESConfiguration(mp4File, mp4Track, (uint8_t **) &escBuffer, (uint32_t *) &escBufferSize);
 
-			ex_NeAACDecInit2(handle, escBuffer, escBufferSize, (unsigned long *) &format.rate, (unsigned char *) &format.channels);
+			if (ex_NeAACDecInit2(handle, escBuffer, escBufferSize, (unsigned long *) &format.rate, (unsigned char *) &format.channels) < 0)
+			{
+				errorState  = True;
+				errorString = "Unsupported audio format";
+			}
 
 			ex_MP4Free(escBuffer);
 
@@ -258,35 +267,42 @@ Error BoCA::DecoderFAAD2::GetStreamInfo(const String &streamURI, Track &track)
 
 		in.InputData((void *) data, size);
 
-		ex_NeAACDecInit(handle, data, size, (unsigned long *) &format.rate, (unsigned char *) &format.channels);
-
-		/* Compute approximate length of stream.
-		 */
-		NeAACDecFrameInfo	 frameInfo;
-
-		ex_NeAACDecDecode(handle, &frameInfo, data, size);
-
-		if (!frameInfo.error)
-		{
-			Int	 bytesConsumed = 0;
-			Int	 samplesRead   = 0;
-
-			while (!frameInfo.error)
-			{
-				bytesConsumed += frameInfo.bytesconsumed;
-				samplesRead   += frameInfo.samples;
-
-				ex_NeAACDecDecode(handle, &frameInfo, data + bytesConsumed, size - bytesConsumed);
-			}
-
-			if (samplesRead > 0) track.approxLength = samplesRead / format.channels * (track.fileSize / bytesConsumed);
-		}
-		else
+		if (ex_NeAACDecInit(handle, data, size, (unsigned long *) &format.rate, (unsigned char *) &format.channels) < 0)
 		{
 			errorState  = True;
 			errorString = "Unsupported audio format";
 		}
 
+		/* Compute approximate length of stream.
+		 */
+		if (!errorState)
+		{
+			NeAACDecFrameInfo	 frameInfo;
+
+			ex_NeAACDecDecode(handle, &frameInfo, data, size);
+
+			if (!frameInfo.error)
+			{
+				Int	 bytesConsumed = 0;
+				Int	 samplesRead   = 0;
+
+				while (!frameInfo.error)
+				{
+					bytesConsumed += frameInfo.bytesconsumed;
+					samplesRead   += frameInfo.samples;
+
+					ex_NeAACDecDecode(handle, &frameInfo, data + bytesConsumed, size - bytesConsumed);
+				}
+
+				if (samplesRead > 0) track.approxLength = samplesRead / format.channels * (track.fileSize / bytesConsumed);
+			}
+			else
+			{
+				errorState  = True;
+				errorString = "Unsupported audio format";
+			}
+		}
+
 		/* Close handles.
 		 */
 		delete [] data;
@@ -358,10 +374,15 @@ Bool BoCA::DecoderFAAD2::Activate()
 
 	if ((in.InputNumberRaw(8) & 0xFFFFFFFF) == 'ftyp')
 	{
-		mp4File	 = ex_MP4Read(track.fileName.ConvertTo("UTF-8"));
+		mp4File	 = ex_MP4ReadProvider(track.fileName.ConvertTo("UTF-8"), NIL);
 		mp4Track = GetAudioTrack(mp4File);
 
-		if (mp4Track == -1) return False;
+		if (mp4Track == -1)
+		{
+			ex_MP4Close(mp4File, 0);
+
+			return False;
+		}
 
 		driver->Seek(0);
 	}
@@ -396,7 +417,11 @@ Bool BoCA::DecoderFAAD2::Activate()
 		unsigned long	 rate;
 		unsigned char	 channels;
 
-		ex_NeAACDecInit2(handle, escBuffer, escBufferSize, &rate, &channels);
+		if (ex_NeAACDecInit2(handle, escBuffer, escBufferSize, &rate, &channels) < 0)
+		{
+			errorState  = True;
+			errorString = "Unsupported audio format";
+		}
 
 		ex_MP4Free(escBuffer);
 	}
@@ -410,13 +435,28 @@ Bool BoCA::DecoderFAAD2::Activate()
 		unsigned long	 rate;
 		unsigned char	 channels;
 
-		ex_NeAACDecInit(handle, data, size, &rate, &channels);
+		if (ex_NeAACDecInit(handle, data, size, &rate, &channels) < 0)
+		{
+			errorState  = True;
+			errorString = "Unsupported audio format";
+		}
 
 		delete [] data;
 
 		driver->Seek(driver->GetPos() - size);
 	}
 
+	/* Check for error.
+	 */
+	if (errorState)
+	{
+		ex_NeAACDecClose(handle);
+
+		if (mp4File != NIL) ex_MP4Close(mp4File, 0);
+
+		return False;
+	}
+
 	return True;
 }
 
