@@ -1,5 +1,5 @@
  /* BoCA - BonkEnc Component Architecture
-  * Copyright (C) 2007-2020 Robert Kausch <robert.kausch@freac.org>
+  * Copyright (C) 2007-2021 Robert Kausch <robert.kausch@freac.org>
   *
   * This program is free software; you can redistribute it and/or
   * modify it under the terms of the GNU General Public License as
@@ -15,6 +15,10 @@
 
 #include "fdkaac.h"
 
+#ifndef CAPF_AAC_USAC
+#	define CAPF_AAC_USAC 0x00200000
+#endif
+
 using namespace smooth::IO;
 
 const String &BoCA::DecoderFDKAAC::GetComponentSpecs()
@@ -27,7 +31,7 @@ const String &BoCA::DecoderFDKAAC::GetComponentSpecs()
 											\
 		  <?xml version=\"1.0\" encoding=\"UTF-8\"?>				\
 		  <component>								\
-		    <name>Fraunhofer FDK AAC Decoder</name>				\
+		    <name>FDK-AAC Decoder %VERSION%</name>				\
 		    <version>1.0</version>						\
 		    <id>fdkaac-dec</id>							\
 		    <type>decoder</type>						\
@@ -46,6 +50,7 @@ const String &BoCA::DecoderFDKAAC::GetComponentSpecs()
 			      <extension>m4r</extension>				\
 			      <extension>mp4</extension>				\
 			      <extension>3gp</extension>				\
+			      <extension>3gpp</extension>				\
 			      <tag id=\"mp4-tag\" mode=\"other\">MP4 Metadata</tag>	\
 			    </format>							\
 											\
@@ -62,6 +67,12 @@ const String &BoCA::DecoderFDKAAC::GetComponentSpecs()
 		  </component>								\
 											\
 		");
+
+		UnsignedInt32	 version = GetDecoderVersion();
+
+		componentSpecs.Replace("%VERSION%", String("v").Append(String::FromInt((version >> 24) & 0xff)).Append(".")
+							       .Append(String::FromInt((version >> 16) & 0xff)).Append(".")
+							       .Append(String::FromInt((version >>  8) & 0xff)));
 	}
 
 	return componentSpecs;
@@ -86,20 +97,31 @@ Bool BoCA::DecoderFDKAAC::CanOpenStream(const String &streamURI)
 
 	if (mp4v2dll != NIL && (in.InputNumberRaw(8) & 0xFFFFFFFF) == 'ftyp')
 	{
-		MP4FileHandle	 mp4File  = ex_MP4Read(streamURI.ConvertTo("UTF-8"));
+		MP4FileHandle	 mp4File  = ex_MP4ReadProvider(streamURI.ConvertTo("UTF-8"), NIL);
 		Int		 mp4Track = GetAudioTrack(mp4File);
 
 		if (mp4Track >= 0 && ex_MP4GetSampleSize(mp4File, mp4Track, 1) > 0)
 		{
-			Int	 type = ex_MP4GetTrackAudioMpeg4Type(mp4File, mp4Track);
+			LIB_INFO	 info[FDK_MODULE_LAST];
 
-			if (type == AOT_AAC_LC	  ||
-			    type == AOT_SBR	  ||
-			    type == AOT_PS	  ||
+			FDKinitLibInfo(info);
+			ex_aacDecoder_GetLibInfo(info);
 
-			    type == AOT_ER_AAC_LC ||
-			    type == AOT_ER_AAC_LD ||
-			    type == AOT_ER_AAC_ELD) isValidFile = True;
+			UINT	 aacFlags = FDKlibInfo_getCapabilities(info, FDK_AACDEC);
+			UINT	 sbrFlags = FDKlibInfo_getCapabilities(info, FDK_SBRDEC);
+
+			Int	 type	  = ex_MP4GetTrackAudioMpeg4Type(mp4File, mp4Track);
+
+			if ((type == AOT_AAC_LC	     && aacFlags & CAPF_AAC_LC	   ) ||
+			    (type == AOT_SBR	     && sbrFlags		   ) ||
+			    (type == AOT_PS	     && sbrFlags & CAPF_SBR_PS_MPEG) ||
+			    (type == AOT_USAC	     && aacFlags & CAPF_AAC_USAC   ) ||
+
+			    (type == AOT_ER_AAC_LC   && aacFlags & CAPF_ER_AAC_LC  ) ||
+			    (type == AOT_ER_AAC_LD   && aacFlags & CAPF_ER_AAC_LD  ) ||
+			    (type == AOT_ER_AAC_ELD  && aacFlags & CAPF_ER_AAC_ELD ) ||
+			    (type == AOT_ER_AAC_SCAL && aacFlags & CAPF_ER_AAC_SCAL) ||
+			    (type == AOT_ER_BSAC     && aacFlags & CAPF_ER_AAC_BSAC)) isValidFile = True;
 		}
 
 		ex_MP4Close(mp4File, 0);
@@ -124,7 +146,7 @@ Error BoCA::DecoderFDKAAC::GetStreamInfo(const String &streamURI, Track &track)
 		track.fileSize	= File(streamURI).GetFileSize();
 		track.length	= -1;
 
-		MP4FileHandle	 mp4File  = ex_MP4Read(streamURI.ConvertTo("UTF-8"));
+		MP4FileHandle	 mp4File  = ex_MP4ReadProvider(streamURI.ConvertTo("UTF-8"), NIL);
 		Int		 mp4Track = GetAudioTrack(mp4File);
 
 		if (mp4Track >= 0 && ex_MP4GetSampleSize(mp4File, mp4Track, 1) > 0)
@@ -366,16 +388,38 @@ BoCA::DecoderFDKAAC::~DecoderFDKAAC()
 {
 }
 
+UnsignedInt32 BoCA::DecoderFDKAAC::GetDecoderVersion()
+{
+	LIB_INFO	 info[FDK_MODULE_LAST];
+
+	FDKinitLibInfo(info);
+	ex_aacDecoder_GetLibInfo(info);
+
+	for (Int i = 0; i < FDK_MODULE_LAST; i++)
+	{
+		if (info[i].module_id != FDK_AACDEC) continue;
+
+		return info[i].version;
+	}
+
+	return 0;
+}
+
 Bool BoCA::DecoderFDKAAC::Activate()
 {
 	InStream	 in(STREAM_DRIVER, driver);
 
 	if ((in.InputNumberRaw(8) & 0xFFFFFFFF) == 'ftyp')
 	{
-		mp4File	 = ex_MP4Read(track.fileName.ConvertTo("UTF-8"));
+		mp4File	 = ex_MP4ReadProvider(track.fileName.ConvertTo("UTF-8"), NIL);
 		mp4Track = GetAudioTrack(mp4File);
 
-		if (mp4Track == -1) return False;
+		if (mp4Track == -1)
+		{
+			ex_MP4Close(mp4File, 0);
+
+			return False;
+		}
 
 		driver->Seek(0);
 
@@ -472,23 +516,21 @@ Int BoCA::DecoderFDKAAC::ReadData(Buffer<UnsignedByte> &data)
 
 			unsigned int	 bytesValid = bufferSize;
 
-			while (True)
+			if (bytesValid > 0)
 			{
-				if (bytesValid > 0)
-				{
-					unsigned char	*inputBuffer	 = buffer + bufferSize - bytesValid;
-					unsigned int	 inputBufferSize = bytesValid;
+				unsigned char	*inputBuffer	 = buffer + bufferSize - bytesValid;
+				unsigned int	 inputBufferSize = bytesValid;
 
-					ex_aacDecoder_Fill(handle, &inputBuffer, &inputBufferSize, &bytesValid);
-				}
+				ex_aacDecoder_Fill(handle, &inputBuffer, &inputBufferSize, &bytesValid);
+			}
 
-				if (frameSize == 0) samplesBuffer.Resize((samplesRead + maxFrameSize) * format.channels);
-				else		    samplesBuffer.Resize((samplesRead + frameSize) * format.channels);
+			if (frameSize == 0) samplesBuffer.Resize((samplesRead + maxFrameSize) * format.channels);
+			else		    samplesBuffer.Resize((samplesRead + frameSize) * format.channels);
 
-				short	*outputBuffer = samplesBuffer + samplesRead * format.channels;
-
-				if (ex_aacDecoder_DecodeFrame(handle, outputBuffer, samplesBuffer.Size() - samplesRead * format.channels, 0) != AAC_DEC_OK) break;
+			short	*outputBuffer = samplesBuffer + samplesRead * format.channels;
 
+			if (ex_aacDecoder_DecodeFrame(handle, outputBuffer, samplesBuffer.Size() - samplesRead * format.channels, 0) == AAC_DEC_OK)
+			{
 				if (frameSize == 0)
 				{
 					CStreamInfo	*streamInfo = ex_aacDecoder_GetStreamInfo(handle);
