@@ -1,5 +1,5 @@
  /* BoCA - BonkEnc Component Architecture
-  * Copyright (C) 2007-2020 Robert Kausch <robert.kausch@freac.org>
+  * Copyright (C) 2007-2021 Robert Kausch <robert.kausch@freac.org>
   *
   * This program is free software; you can redistribute it and/or
   * modify it under the terms of the GNU General Public License as
@@ -17,6 +17,7 @@
 #	include <windows.h>
 #else
 #	include <sys/mman.h>
+#	include <sys/sem.h>
 #	include <sys/wait.h>
 #	include <stdlib.h>
 #	include <unistd.h>
@@ -34,6 +35,15 @@ const String &BoCA::EncoderCoreAudioConnect::GetComponentSpecs()
 {
 	static String	 componentSpecs;
 
+#ifndef __WIN32__
+	const Array<String>	&args = GUI::Application::GetArguments();
+
+	foreach (const String &arg, args)
+	{
+		if (arg == "--disable-wine") return componentSpecs;
+	}
+#endif
+
 	if (EncoderCoreAudioConnect().IsReady())
 	{
 		componentSpecs = "									\
@@ -568,27 +578,39 @@ Bool BoCA::EncoderCoreAudioConnect::ConvertArguments(Config *config)
 
 	static const String	 encoderID = "coreaudio-enc";
 
+	/* Set default values.
+	 */
+	if (!config->GetIntValue("Settings", "UserSpecifiedConfig", False))
+	{
+		config->SetIntValue(ConfigureCoreAudio::ConfigID, "MP4Container", True);
+
+		config->SetIntValue(ConfigureCoreAudio::ConfigID, "Codec", CA::kAudioFormatMPEG4AAC);
+		config->SetIntValue(ConfigureCoreAudio::ConfigID, "Bitrate", 64);
+	}
+
 	/* Get command line settings.
 	 */
-	Int	 bitrate = 64;
-	String	 format	 = "AAC";
+	Bool	 rawAAC	    = config->GetIntValue(encoderID, "Write raw AAC files", !config->GetIntValue(ConfigureCoreAudio::ConfigID, "MP4Container", True));
 
-	if (config->GetIntValue(encoderID, "Set Bitrate per channel", False)) bitrate = config->GetIntValue(encoderID, "Bitrate per channel", bitrate);
-	if (config->GetIntValue(encoderID, "Set Output format", False))	      format  = config->GetStringValue(encoderID, "Output format", format).ToUpper();
+	Int	 bitrate    = config->GetIntValue(ConfigureCoreAudio::ConfigID, "Bitrate", 64);
+	Int	 format	    = config->GetIntValue(ConfigureCoreAudio::ConfigID, "Codec", CA::kAudioFormatMPEG4AAC);
+	String	 formatName = "AAC";
+
+	if (format == CA::kAudioFormatAppleLossless) formatName = "ALAC";
+
+	if (config->GetIntValue(encoderID, "Set Bitrate per channel", False)) bitrate	 = config->GetIntValue(encoderID, "Bitrate per channel", bitrate);
+	if (config->GetIntValue(encoderID, "Set Output format", False))	      formatName = config->GetStringValue(encoderID, "Output format", formatName).ToUpper();
 
 	/* Set configuration values.
 	 */
-	config->SetIntValue(ConfigureCoreAudio::ConfigID, "MP4Container", !config->GetIntValue(encoderID, "Write raw AAC files", False) || format == "ALAC");
+	config->SetIntValue(ConfigureCoreAudio::ConfigID, "MP4Container", !rawAAC || formatName == "ALAC");
 
+	if	(formatName == "AAC" ) format = CA::kAudioFormatMPEG4AAC;
+	else if (formatName == "ALAC") format = CA::kAudioFormatAppleLossless;
+
+	config->SetIntValue(ConfigureCoreAudio::ConfigID, "Codec", format);
 	config->SetIntValue(ConfigureCoreAudio::ConfigID, "Bitrate", Math::Max(8, Math::Min(256, bitrate)));
 
-	Int	 audioFormat = CA::kAudioFormatMPEG4AAC;
-
-	if	(format == "AAC" ) audioFormat = CA::kAudioFormatMPEG4AAC;
-	else if (format == "ALAC") audioFormat = CA::kAudioFormatAppleLossless;
-
-	config->SetIntValue(ConfigureCoreAudio::ConfigID, "Codec", audioFormat);
-
 	return True;
 }
 
@@ -651,15 +673,20 @@ Bool BoCA::EncoderCoreAudioConnect::Connect()
 		/* Create shared memory object and map view to communication buffer.
 		 */
 #ifdef __WIN32__
-		mappingName = String("freac:").Append(Number((Int64) GetCurrentProcessId()).ToHexString()).Append("-").Append(Number((Int64) GetCurrentThreadId()).ToHexString()).Append("-").Append(Number((Int64) count++).ToHexString());
+		mappingName = BoCA::GetApplicationPrefix().Append(":").Append(Number((Int64) GetCurrentProcessId()).ToHexString()).Append("-").Append(Number((Int64) GetCurrentThreadId()).ToHexString()).Append("-").Append(Number((Int64) count++).ToHexString());
 
+		semaphore   = CreateSemaphoreA(NIL, 0, 1, String(mappingName).Append("-sem"));
 		mapping	    = CreateFileMappingA(INVALID_HANDLE_VALUE, NIL, PAGE_READWRITE, 0, sizeof(CoreAudioCommBuffer), mappingName);
 		comm	    = (CoreAudioCommBuffer *) MapViewOfFile(mapping, FILE_MAP_ALL_ACCESS, 0, 0, 0);
 #else
-		mappingName = String("/freac:").Append(Number((Int64) getpid()).ToHexString()).Append("-").Append(Number((Int64) pthread_self()).ToHexString()).Append("-").Append(Number((Int64) count++).ToHexString());
+		mappingName = String("/").Append(BoCA::GetApplicationPrefix()).Append(":").Append(Number((Int64) getpid()).ToHexString()).Append("-").Append(Number((Int64) pthread_self()).ToHexString()).Append("-").Append(Number((Int64) count++).ToHexString());
 
+		key_t	 key = mappingName.ComputeCRC32();
+
+		semaphore   = semget(key, 1, IPC_CREAT | IPC_EXCL | 0666);
 		mapping	    = shm_open(mappingName, O_CREAT | O_EXCL | O_RDWR, 0666);
 
+		semctl(semaphore, 0, SETVAL, 0);
 		ftruncate(mapping, sizeof(CoreAudioCommBuffer));
 
 		comm	    = (CoreAudioCommBuffer *) mmap(NULL, sizeof(CoreAudioCommBuffer), PROT_READ | PROT_WRITE, MAP_SHARED, mapping, 0);
@@ -684,7 +711,7 @@ Bool BoCA::EncoderCoreAudioConnect::Connect()
 
 		connector = execInfo.hProcess;
 #else
-		const char	*cmd = String("wine ").Append(file).Append(" ").Append(mappingName).Append(" 2> /dev/null");
+		const char	*cmd = String("wine ").Append(file).Append(" ").Append(mappingName).Append(" ").Append(String::FromInt(key)).Append(" 2> /dev/null");
 
 		connector = fork();
 
@@ -737,11 +764,13 @@ Bool BoCA::EncoderCoreAudioConnect::Disconnect()
 #ifdef __WIN32__
 	UnmapViewOfFile(comm);
 	CloseHandle(mapping);
+	CloseHandle(semaphore);
 #else
 	munmap(comm, sizeof(CoreAudioCommBuffer));
 	close(mapping);
 
 	shm_unlink(mappingName);
+	semctl(semaphore, 0, IPC_RMID, 0);
 #endif
 
 	connected = False;
@@ -753,23 +782,37 @@ Bool BoCA::EncoderCoreAudioConnect::ProcessConnectorCommand()
 {
 	if (!connected) return False;
 
-	comm->status = CommStatusIssued;
-
-	while (comm->status == CommStatusIssued || comm->status == CommStatusProcessing)
-	{
-		/* Check if the connector still exists.
-		 */
-#ifdef __WIN32__
-		if (WaitForSingleObject(connector, 0) != WAIT_TIMEOUT) break;
-#else
-		int	 status = 0;
-
-		if (waitpid(connector, &status, WNOHANG) != 0) break;
+#ifndef __WIN32__
+	static sembuf	 opWait[1] = { 0, -1, SEM_UNDO };
+	static sembuf	 opPost[1] = { 0, 1,  SEM_UNDO };
 #endif
 
-		/* Sleep for one millisecond.
+	comm->status = CommStatusIssued;
+
+	while (True)
+	{
+#ifdef __WIN32__
+		ReleaseSemaphore(semaphore, 1, NULL);
+
+		while (WaitForSingleObject(semaphore, 1000) != WAIT_OBJECT_0)
+		{
+			/* Check if the connector still exists.
+			 */
+			if (WaitForSingleObject(connector, 0) != WAIT_TIMEOUT) return False;
+		}
+#else
+		semop(semaphore, opPost, 1);
+
+		if (semop(semaphore, opWait, 1) != 0) return False;
+
+		/* Check if the connector still exists.
 		 */
-		S::System::System::Sleep(1);
+		int	 status = 0;
+
+		if (waitpid(connector, &status, WNOHANG) != 0) return False;
+#endif
+
+		if (comm->status != CommStatusIssued) break;
 	}
 
 	return True;
