@@ -1,5 +1,5 @@
  /* BoCA - BonkEnc Component Architecture
-  * Copyright (C) 2007-2020 Robert Kausch <robert.kausch@freac.org>
+  * Copyright (C) 2007-2021 Robert Kausch <robert.kausch@freac.org>
   *
   * This program is free software; you can redistribute it and/or
   * modify it under the terms of the GNU General Public License as
@@ -22,11 +22,18 @@ const String &BoCA::EncoderFDKAAC::GetComponentSpecs()
 
 	if (fdkaacdll != NIL)
 	{
+		LIB_INFO	 info[FDK_MODULE_LAST];
+
+		FDKinitLibInfo(info);
+		ex_aacEncGetLibInfo(info);
+
+		UINT	 sbrFlags = FDKlibInfo_getCapabilities(info, FDK_SBRENC);
+
 		componentSpecs = "										\
 														\
 		  <?xml version=\"1.0\" encoding=\"UTF-8\"?>							\
 		  <component>											\
-		    <name>Fraunhofer FDK AAC Encoder</name>							\
+		    <name>FDK-AAC Encoder %VERSION%</name>							\
 		    <version>1.0</version>									\
 		    <id>fdkaac-enc</id>										\
 		    <type>encoder</type>									\
@@ -63,10 +70,22 @@ const String &BoCA::EncoderFDKAAC::GetComponentSpecs()
 		    <parameters>										\
 		      <selection name=\"AAC encoding mode\" argument=\"-m %VALUE\" default=\"LC\">		\
 			<option alias=\"Low Complexity\">LC</option>						\
-			<option alias=\"High Efficiency\">HE</option>						\
-			<option alias=\"High Efficiency v2\">HEv2</option>					\
+														\
+		");
+
+		if (sbrFlags)			 componentSpecs.Append("<option alias=\"High Efficiency\">HE</option>");
+		if (sbrFlags & CAPF_SBR_PS_MPEG) componentSpecs.Append("<option alias=\"High Efficiency v2\">HEv2</option>");
+
+		componentSpecs.Append("										\
+														\
 			<option alias=\"Low Delay\">LD</option>							\
-			<option alias=\"Enhanced Low Delay\">ELD</option>					\
+														\
+		");
+
+		if (sbrFlags)			 componentSpecs.Append("<option alias=\"Enhanced Low Delay\">ELD</option>");
+
+		componentSpecs.Append("										\
+														\
 		      </selection>										\
 		      <range name=\"Bitrate per channel\" argument=\"-b %VALUE\" default=\"64\">		\
 			<min alias=\"min\">8</min>								\
@@ -77,6 +96,12 @@ const String &BoCA::EncoderFDKAAC::GetComponentSpecs()
 		  </component>											\
 														\
 		");
+
+		UnsignedInt32	 version = GetEncoderVersion();
+
+		componentSpecs.Replace("%VERSION%", String("v").Append(String::FromInt((version >> 24) & 0xff)).Append(".")
+							       .Append(String::FromInt((version >> 16) & 0xff)).Append(".")
+							       .Append(String::FromInt((version >>  8) & 0xff)));
 	}
 
 	return componentSpecs;
@@ -125,13 +150,12 @@ BoCA::EncoderFDKAAC::~EncoderFDKAAC()
 
 UnsignedInt32 BoCA::EncoderFDKAAC::GetEncoderVersion()
 {
-	LIB_INFO	 info[0xFF];
-
-	memset(info, 0, sizeof(info));
+	LIB_INFO	 info[FDK_MODULE_LAST];
 
+	FDKinitLibInfo(info);
 	ex_aacEncGetLibInfo(info);
 
-	for (int i = 0; i < 0xFF; i++)
+	for (Int i = 0; i < FDK_MODULE_LAST; i++)
 	{
 		if (info[i].module_id != FDK_AACENC) continue;
 
@@ -154,7 +178,9 @@ Bool BoCA::EncoderFDKAAC::Activate()
 	Bool	 mp4Container = config->GetIntValue(ConfigureFDKAAC::ConfigID, "MP4Container", True);
 	Int	 mpegVersion  = config->GetIntValue(ConfigureFDKAAC::ConfigID, "MPEGVersion", 0);
 	Int	 aacType      = config->GetIntValue(ConfigureFDKAAC::ConfigID, "AACType", AOT_AAC_LC);
+	Int	 mode	      = config->GetIntValue(ConfigureFDKAAC::ConfigID, "Mode", 0);
 	Int	 bitrate      = config->GetIntValue(ConfigureFDKAAC::ConfigID, "Bitrate", 64);
+	Int	 quality      = config->GetIntValue(ConfigureFDKAAC::ConfigID, "Quality", 4);
 
 	/* Create and configure FDK AAC encoder.
 	 */
@@ -178,15 +204,14 @@ Bool BoCA::EncoderFDKAAC::Activate()
 	ex_aacEncoder_SetParam(handle, AACENC_SAMPLERATE, format.rate);
 	ex_aacEncoder_SetParam(handle, AACENC_CHANNELMODE, channelMode);
 
+	if (mode == 0) ex_aacEncoder_SetParam(handle, AACENC_BITRATE, bitrate * 1000 * format.channels);
+	else	       ex_aacEncoder_SetParam(handle, AACENC_BITRATEMODE, quality);
+
 	ex_aacEncoder_SetParam(handle, AACENC_AOT, mpegVersion + aacType);
-	ex_aacEncoder_SetParam(handle, AACENC_BITRATE, bitrate * 1000 * format.channels);
 	ex_aacEncoder_SetParam(handle, AACENC_TRANSMUX, mp4Container ? TT_MP4_RAW : TT_MP4_ADTS);
 
-	if (!mp4Container)
-	{
-		if (aacType == AOT_ER_AAC_LD ||
-		    aacType == AOT_ER_AAC_ELD) ex_aacEncoder_SetParam(handle, AACENC_TRANSMUX, TT_MP4_LOAS);
-	}
+	if ( mp4Container && (aacType == AOT_SBR       || aacType == AOT_PS	   )) ex_aacEncoder_SetParam(handle, AACENC_SIGNALING_MODE, 1);
+	if (!mp4Container && (aacType == AOT_ER_AAC_LD || aacType == AOT_ER_AAC_ELD)) ex_aacEncoder_SetParam(handle, AACENC_TRANSMUX, TT_MP4_LOAS);
 
 	AACENC_InfoStruct	 aacInfo;
 	AACENC_InfoStruct_Old	&aacInfoOld = *(AACENC_InfoStruct_Old *) &aacInfo;
@@ -198,7 +223,7 @@ Bool BoCA::EncoderFDKAAC::Activate()
 	granuleSize  = ex_aacEncoder_GetParam(handle, AACENC_GRANULE_LENGTH);
 
 #if AACENCODER_LIB_VL0 >= 4
-	if (GetEncoderVersion() >= 4 << 24)
+	if (GetEncoderVersion() >= LIB_VERSION(4, 0, 0))
 	{
 		delaySamples = aacInfo.nDelayCore;
 	}
@@ -231,8 +256,8 @@ Bool BoCA::EncoderFDKAAC::Activate()
 
 		ex_MP4SetAudioProfileLevel(mp4File, 0x0F);
 
-		if (GetEncoderVersion() >= 4 << 24) ex_MP4SetTrackESConfiguration(mp4File, mp4Track, aacInfo.confBuf, aacInfo.confSize);
-		else				    ex_MP4SetTrackESConfiguration(mp4File, mp4Track, aacInfoOld.confBuf, aacInfoOld.confSize);
+		if (GetEncoderVersion() >= LIB_VERSION(4, 0, 0)) ex_MP4SetTrackESConfiguration(mp4File, mp4Track, aacInfo.confBuf, aacInfo.confSize);
+		else						 ex_MP4SetTrackESConfiguration(mp4File, mp4Track, aacInfoOld.confBuf, aacInfoOld.confSize);
 
 		totalSamples = 0;
 	}
@@ -453,9 +478,7 @@ Int BoCA::EncoderFDKAAC::EncodeFrames(Bool flush)
 	{
 		SuperWorker	*workerToUse = workers.GetNth(nextWorker % workers.Length());
 
-		while (!workerToUse->IsReady()) S::System::System::Sleep(1);
-
-		workerToUse->Lock();
+		workerToUse->WaitUntilReady();
 
 		/* See if the worker has some packets for us.
 		 */
@@ -464,7 +487,6 @@ Int BoCA::EncoderFDKAAC::EncodeFrames(Bool flush)
 		/* Pass new frames to worker.
 		 */
 		workerToUse->Encode(samplesBuffer, framesProcessed * samplesPerFrame, flush ? samplesBuffer.Size() : samplesPerFrame * framesToProcess, flush);
-		workerToUse->Release();
 
 		framesProcessed += framesToProcess - (flush ? 0 : overlap);
 
@@ -485,16 +507,12 @@ Int BoCA::EncoderFDKAAC::EncodeFrames(Bool flush)
 	{
 		SuperWorker	*workerToUse = workers.GetNth(nextWorker % workers.Length());
 
-		while (!workerToUse->IsReady()) S::System::System::Sleep(1);
-
-		workerToUse->Lock();
+		workerToUse->WaitUntilReady();
 
 		/* See if the worker has some packets for us.
 		 */
 		if (workerToUse->GetPacketSizes().Length() != 0) dataLength += ProcessPackets(workerToUse->GetPackets(), workerToUse->GetPacketSizes(), nextWorker == workers.Length());
 
-		workerToUse->Release();
-
 		nextWorker++;
 	}
 
@@ -565,31 +583,45 @@ Bool BoCA::EncoderFDKAAC::ConvertArguments(Config *config)
 
 	static const String	 encoderID = "fdkaac-enc";
 
+	/* Set default values.
+	 */
+	if (!config->GetIntValue("Settings", "UserSpecifiedConfig", False))
+	{
+		config->SetIntValue(ConfigureFDKAAC::ConfigID, "MPEGVersion", 0);
+		config->SetIntValue(ConfigureFDKAAC::ConfigID, "MP4Container", True);
+
+		config->SetIntValue(ConfigureFDKAAC::ConfigID, "AACType", AOT_AAC_LC);
+		config->SetIntValue(ConfigureFDKAAC::ConfigID, "Bitrate", 64);
+	}
+
 	/* Get command line settings.
 	 */
-	Int	 bitrate = 64;
-	String	 mode	 = "LC";
+	Bool	 rawAAC	       = config->GetIntValue(encoderID, "Write raw AAC files", !config->GetIntValue(ConfigureFDKAAC::ConfigID, "MP4Container", True));
 
-	if (config->GetIntValue(encoderID, "Set Bitrate per channel", False)) bitrate = config->GetIntValue(encoderID, "Bitrate per channel", bitrate);
-	if (config->GetIntValue(encoderID, "Set AAC encoding mode", False))   mode    = config->GetStringValue(encoderID, "AAC encoding mode", mode).ToUpper();
+	Int	 bitrate       = config->GetIntValue(ConfigureFDKAAC::ConfigID, "Bitrate", 64);
+	Int	 aacType       = config->GetIntValue(ConfigureFDKAAC::ConfigID, "AACType", AOT_AAC_LC);
+	String	 aacTypeString = "LC";
+
+	if	(aacType == AOT_SBR)	    aacTypeString = "HE";
+	else if	(aacType == AOT_PS)	    aacTypeString = "HEv2";
+	else if	(aacType == AOT_ER_AAC_LD)  aacTypeString = "LD";
+	else if	(aacType == AOT_ER_AAC_ELD) aacTypeString = "ELD";
+
+	if (config->GetIntValue(encoderID, "Set Bitrate per channel", False)) bitrate	    = config->GetIntValue(encoderID, "Bitrate per channel", bitrate);
+	if (config->GetIntValue(encoderID, "Set AAC encoding mode", False))   aacTypeString = config->GetStringValue(encoderID, "AAC encoding mode", aacTypeString).ToUpper();
 
 	/* Set configuration values.
 	 */
-	config->SetIntValue(ConfigureFDKAAC::ConfigID, "MPEGVersion", 0);
+	config->SetIntValue(ConfigureFDKAAC::ConfigID, "MP4Container", !rawAAC);
 
-	config->SetIntValue(ConfigureFDKAAC::ConfigID, "MP4Container", !config->GetIntValue(encoderID, "Write raw AAC files", False));
-
-	config->SetIntValue(ConfigureFDKAAC::ConfigID, "Bitrate", Math::Max(8, Math::Min(256, bitrate)));
-
-	Int	 aacType = AOT_AAC_LC;
-
-	if	(mode == "LC"  ) aacType = AOT_AAC_LC;
-	else if (mode == "HE"  ) aacType = AOT_SBR;
-	else if (mode == "HEV2") aacType = AOT_PS;
-	else if (mode == "LD"  ) aacType = AOT_ER_AAC_LD;
-	else if (mode == "ELD" ) aacType = AOT_ER_AAC_ELD;
+	if	(aacTypeString == "LC"  ) aacType = AOT_AAC_LC;
+	else if (aacTypeString == "HE"  ) aacType = AOT_SBR;
+	else if (aacTypeString == "HEV2") aacType = AOT_PS;
+	else if (aacTypeString == "LD"  ) aacType = AOT_ER_AAC_LD;
+	else if (aacTypeString == "ELD" ) aacType = AOT_ER_AAC_ELD;
 
 	config->SetIntValue(ConfigureFDKAAC::ConfigID, "AACType", aacType);
+	config->SetIntValue(ConfigureFDKAAC::ConfigID, "Bitrate", Math::Max(8, Math::Min(256, bitrate)));
 
 	return True;
 }
