@@ -1,5 +1,5 @@
  /* BoCA - BonkEnc Component Architecture
-  * Copyright (C) 2007-2020 Robert Kausch <robert.kausch@freac.org>
+  * Copyright (C) 2007-2021 Robert Kausch <robert.kausch@freac.org>
   *
   * This program is free software; you can redistribute it and/or
   * modify it under the terms of the GNU General Public License as
@@ -90,13 +90,13 @@ Error BoCA::TaggerAPEv2::RenderBuffer(Buffer<UnsignedByte> &buffer, const Track
 
 	Int		 numItems = 0;
 
-	if (info.artist != NIL) { RenderAPEItem("Artist", info.artist, buffer);		     numItems++; }
-	if (info.title  != NIL) { RenderAPEItem("Title", info.title, buffer);		     numItems++; }
-	if (info.album  != NIL) { RenderAPEItem("Album", info.album, buffer);		     numItems++; }
-	if (info.year    >   0) { RenderAPEItem("Year", String::FromInt(info.year), buffer); numItems++; }
-	if (info.genre  != NIL) { RenderAPEItem("Genre", info.genre, buffer);		     numItems++; }
-	if (info.label  != NIL) { RenderAPEItem("Publisher", info.label, buffer);	     numItems++; }
-	if (info.isrc   != NIL) { RenderAPEItem("ISRC", info.isrc, buffer);		     numItems++; }
+	if (info.artist != NIL) { RenderAPEItem("Artist",    info.artist,		 buffer); numItems++; }
+	if (info.title  != NIL) { RenderAPEItem("Title",     info.title,		 buffer); numItems++; }
+	if (info.album  != NIL) { RenderAPEItem("Album",     info.album,		 buffer); numItems++; }
+	if (info.year    >   0) { RenderAPEItem("Year",	     String::FromInt(info.year), buffer); numItems++; }
+	if (info.genre  != NIL) { RenderAPEItem("Genre",     info.genre,		 buffer); numItems++; }
+	if (info.label  != NIL) { RenderAPEItem("Publisher", info.label,		 buffer); numItems++; }
+	if (info.isrc   != NIL) { RenderAPEItem("ISRC",	     info.isrc,			 buffer); numItems++; }
 
 	if (info.track > 0)
 	{
@@ -116,8 +116,8 @@ Error BoCA::TaggerAPEv2::RenderBuffer(Buffer<UnsignedByte> &buffer, const Track
 		{ RenderAPEItem("Disc", discString, buffer); numItems++; }
 	}
 
-	if	(info.comment != NIL && !replaceExistingComments) { RenderAPEItem("Comment", info.comment, buffer);   numItems++; }
-	else if (defaultComment != NIL && numItems > 0)		  { RenderAPEItem("Comment", defaultComment, buffer); numItems++; }
+	if	(info.comment != NIL && !replaceExistingComments) { RenderAPEItem("Comment", info.comment,   buffer, False); numItems++; }
+	else if (defaultComment != NIL && numItems > 0)		  { RenderAPEItem("Comment", defaultComment, buffer	  ); numItems++; }
 
 	/* Save other text info.
 	 */
@@ -128,30 +128,44 @@ Error BoCA::TaggerAPEv2::RenderBuffer(Buffer<UnsignedByte> &buffer, const Track
 
 		if (value == NIL) continue;
 
-		if	(key == INFO_ALBUMARTIST)    { RenderAPEItem("Album Artist",   value, buffer); numItems++; }
+		if	(key == INFO_ALBUMARTIST)    { RenderAPEItem("Album Artist",   value, buffer	   ); numItems++; }
 
-		else if	(key == INFO_CONTENTGROUP)   { RenderAPEItem("Grouping",       value, buffer); numItems++; }
-		else if	(key == INFO_SUBTITLE)	     { RenderAPEItem("Subtitle",       value, buffer); numItems++; }
+		else if	(key == INFO_CONTENTGROUP)   { RenderAPEItem("Grouping",       value, buffer	   ); numItems++; }
+		else if	(key == INFO_SUBTITLE)	     { RenderAPEItem("Subtitle",       value, buffer	   ); numItems++; }
 
-		else if	(key == INFO_PERFORMER)	     { RenderAPEItem("Performer",      value, buffer); numItems++; }
-		else if	(key == INFO_CONDUCTOR)	     { RenderAPEItem("Conductor",      value, buffer); numItems++; }
-		else if	(key == INFO_REMIXER)	     { RenderAPEItem("MixArtist",      value, buffer); numItems++; }
-		else if	(key == INFO_COMPOSER)	     { RenderAPEItem("Composer",       value, buffer); numItems++; }
-		else if	(key == INFO_ARRANGER)	     { RenderAPEItem("Arranger",       value, buffer); numItems++; }
-		else if	(key == INFO_PRODUCER)	     { RenderAPEItem("Producer",       value, buffer); numItems++; }
-		else if	(key == INFO_ENGINEER)	     { RenderAPEItem("Engineer",       value, buffer); numItems++; }
+		else if	(key == INFO_BAND)	     { RenderAPEItem("Ensemble",       value, buffer	   ); numItems++; }
+		else if	(key == INFO_PERFORMER)	     { RenderAPEItem("Performer",      value, buffer	   ); numItems++; }
+		else if	(key == INFO_CONDUCTOR)	     { RenderAPEItem("Conductor",      value, buffer	   ); numItems++; }
+		else if	(key == INFO_REMIXER)	     { RenderAPEItem("MixArtist",      value, buffer	   ); numItems++; }
+		else if	(key == INFO_COMPOSER)	     { RenderAPEItem("Composer",       value, buffer	   ); numItems++; }
+		else if	(key == INFO_ARRANGER)	     { RenderAPEItem("Arranger",       value, buffer	   ); numItems++; }
+		else if	(key == INFO_PRODUCER)	     { RenderAPEItem("Producer",       value, buffer	   ); numItems++; }
+		else if	(key == INFO_ENGINEER)	     { RenderAPEItem("Engineer",       value, buffer	   ); numItems++; }
 
-		else if	(key == INFO_BPM)	     { RenderAPEItem("BPM",	       value, buffer); numItems++; }
-		else if	(key == INFO_INITIALKEY)     { RenderAPEItem("InitialKey",     value, buffer); numItems++; }
+		else if	(key == INFO_MOVEMENT)	     { RenderAPEItem("Movement",       value, buffer	   ); numItems++; }
+		else if	(key == INFO_MOVEMENTTOTAL)  { RenderAPEItem("MovementTotal",  value, buffer	   ); numItems++; }
+		else if	(key == INFO_MOVEMENTNAME)   { RenderAPEItem("MovementName",   value, buffer	   ); numItems++; }
 
-		else if	(key == INFO_COPYRIGHT)	     { RenderAPEItem("Copyright",      value, buffer); numItems++; }
-		else if	(key == INFO_CATALOGNUMBER)  { RenderAPEItem("CatalogNumber",  value, buffer); numItems++; }
+		else if	(key == INFO_BPM)	     { RenderAPEItem("BPM",	       value, buffer	   ); numItems++; }
+		else if	(key == INFO_INITIALKEY)     { RenderAPEItem("InitialKey",     value, buffer	   ); numItems++; }
 
-		else if	(key == INFO_RELEASECOUNTRY) { RenderAPEItem("ReleaseCountry", value, buffer); numItems++; }
+		else if	(key == INFO_COPYRIGHT)	     { RenderAPEItem("Copyright",      value, buffer	   ); numItems++; }
 
-		else if	(key == INFO_DISCSUBTITLE)   { RenderAPEItem("DiscSubtitle",   value, buffer); numItems++; }
+		else if	(key == INFO_MEDIATYPE)	     { RenderAPEItem("Media",	       value, buffer	   ); numItems++; }
+		else if	(key == INFO_CATALOGNUMBER)  { RenderAPEItem("CatalogNumber",  value, buffer	   ); numItems++; }
+		else if	(key == INFO_BARCODE)	     { RenderAPEItem("Barcode",	       value, buffer	   ); numItems++; }
 
-		else if	(key == INFO_WEB_ARTIST)     { RenderAPEItem("Weblink",	       value, buffer); numItems++; }
+		else if	(key == INFO_RELEASECOUNTRY) { RenderAPEItem("ReleaseCountry", value, buffer	   ); numItems++; }
+
+		else if	(key == INFO_DISCSUBTITLE)   { RenderAPEItem("DiscSubtitle",   value, buffer	   ); numItems++; }
+
+		else if	(key == INFO_LYRICS)	     { RenderAPEItem("Lyrics",	       value, buffer, False); numItems++; }
+
+		else if	(key == INFO_WEB_ARTIST)     { RenderAPEItem("Artist URL",     value, buffer	   ); numItems++; }
+		else if	(key == INFO_WEB_PUBLISHER)  { RenderAPEItem("Publisher URL",  value, buffer	   ); numItems++; }
+		else if	(key == INFO_WEB_SOURCE)     { RenderAPEItem("File URL",       value, buffer	   ); numItems++; }
+		else if	(key == INFO_WEB_COPYRIGHT)  { RenderAPEItem("Copyright URL",  value, buffer	   ); numItems++; }
+		else if	(key == INFO_WEB_COMMERCIAL) { RenderAPEItem("Buy URL",	       value, buffer	   ); numItems++; }
 	}
 
 	/* Save Replay Gain info.
@@ -173,10 +187,14 @@ Error BoCA::TaggerAPEv2::RenderBuffer(Buffer<UnsignedByte> &buffer, const Track
 
 	/* Save CD table of contents.
 	 */
-	if (writeMCDI)
-	{
-		if (info.mcdi.GetData().Size() > 0) { RenderAPEBinaryItem("MCDI", info.mcdi.GetData(), buffer); numItems++; }
-	}
+	if (writeMCDI && info.mcdi.IsValid()) { RenderAPEBinaryItem("MCDI", info.mcdi.GetData(), buffer); numItems++; }
+
+	/* Save encoder version.
+	 */
+	Application	*app = Application::Get();
+
+	{ RenderAPEItem("Tool Name",	app->getClientName.Call(),    buffer); numItems++; }
+	{ RenderAPEItem("Tool Version",	app->getClientVersion.Call(), buffer); numItems++; }
 
 	/* Save cover art.
 	 */
@@ -247,7 +265,7 @@ Error BoCA::TaggerAPEv2::RenderBuffer(Buffer<UnsignedByte> &buffer, const Track
 
 			/* Create temporary cuesheet file.
 			 */
-			String	 cueFile  = S::System::System::GetTempDirectory().Append("cuesheet_temp_").Append(String::FromInt(S::System::System::Clock())).Append(".cue");
+			String	 cueFile  = S::System::System::GetTempDirectory().Append("cuesheet_temp_").Append(String::FromInt(track.fileName.ComputeCRC32())).Append(".cue");
 
 			cuesheet->SetTrackList(chapters);
 			cuesheet->WritePlaylist(cueFile);
@@ -384,6 +402,7 @@ Error BoCA::TaggerAPEv2::ParseBuffer(const Buffer<UnsignedByte> &buffer, Track &
 		String			 id;
 		String			 value;
 		Buffer<UnsignedByte>	 item;
+		Int			 previousOffset = offset;
 
 		ParseAPEItem(buffer, offset, &id, &value);
 
@@ -397,16 +416,30 @@ Error BoCA::TaggerAPEv2::ParseBuffer(const Buffer<UnsignedByte> &buffer, Track &
 		else if (id == "ALBUM")		 info.album   = value;
 		else if (id == "YEAR")		 info.year    = value.ToInt();
 		else if (id == "GENRE")		 info.genre   = value;
-		else if (id == "COMMENT")	 info.comment = value;
+
+		else if (id == "COMMENT")
+		{
+			ParseAPEItem(buffer, previousOffset, &id, &value, False);
+
+			info.comment = value;
+		}
+
 		else if (id == "PUBLISHER")	 info.label   = value;
 		else if (id == "LABEL")		 info.label   = value;
-		else if (id == "ISRC")		 info.isrc    = value;
+
+		else if (id == "ISRC")
+		{
+			if (Info::IsISRC(value)) info.isrc = value;
+		}
 
 		else if (id == "ALBUM ARTIST")	 info.SetOtherInfo(INFO_ALBUMARTIST,	value);
 
 		else if (id == "GROUPING")	 info.SetOtherInfo(INFO_CONTENTGROUP,	value);
 		else if (id == "SUBTITLE")	 info.SetOtherInfo(INFO_SUBTITLE,	value);
 
+		else if (id == "ENSEMBLE" ||
+			 id == "ORCHESTRA")	 info.SetOtherInfo(INFO_BAND,		value);
+
 		else if (id == "PERFORMER")	 info.SetOtherInfo(INFO_PERFORMER,	value);
 		else if (id == "CONDUCTOR")	 info.SetOtherInfo(INFO_CONDUCTOR,	value);
 		else if (id == "MIXARTIST")	 info.SetOtherInfo(INFO_REMIXER,	value);
@@ -415,17 +448,47 @@ Error BoCA::TaggerAPEv2::ParseBuffer(const Buffer<UnsignedByte> &buffer, Track &
 		else if (id == "PRODUCER")	 info.SetOtherInfo(INFO_PRODUCER,	value);
 		else if (id == "ENGINEER")	 info.SetOtherInfo(INFO_ENGINEER,	value);
 
-		else if (id == "BPM")		 info.SetOtherInfo(INFO_BPM,		value);
+		else if (id == "MOVEMENT")
+		{
+			info.SetOtherInfo(INFO_MOVEMENT, String::FromInt(value.ToInt()));
+
+			if (value.Contains("/")) info.SetOtherInfo(INFO_MOVEMENTTOTAL, String::FromInt(value.Tail(value.Length() - value.Find("/") - 1).ToInt()));
+		}
+
+		else if (id == "MOVEMENTTOTAL")	 info.SetOtherInfo(INFO_MOVEMENTTOTAL,	value);
+		else if (id == "MOVEMENTNAME")	 info.SetOtherInfo(INFO_MOVEMENTNAME,	value);
+
+		else if (id == "BPM")
+		{
+			if (value.ToInt() > 0) info.SetOtherInfo(INFO_BPM, value);
+		}
+
 		else if (id == "INITIALKEY")	 info.SetOtherInfo(INFO_INITIALKEY,	value);
 
 		else if (id == "COPYRIGHT")	 info.SetOtherInfo(INFO_COPYRIGHT,	value);
+
+		else if (id == "MEDIA")		 info.SetOtherInfo(INFO_MEDIATYPE,	value);
 		else if (id == "CATALOGNUMBER")  info.SetOtherInfo(INFO_CATALOGNUMBER,	value);
+		else if (id == "BARCODE")	 info.SetOtherInfo(INFO_BARCODE,	value);
 
 		else if (id == "RELEASECOUNTRY") info.SetOtherInfo(INFO_RELEASECOUNTRY,	value);
 
 		else if (id == "DISCSUBTITLE")	 info.SetOtherInfo(INFO_DISCSUBTITLE,	value);
 
-		else if (id == "WEBLINK")	 info.SetOtherInfo(INFO_WEB_ARTIST,	value);
+		else if (id == "LYRICS")
+		{
+			ParseAPEItem(buffer, previousOffset, &id, &value, False);
+
+			info.SetOtherInfo(INFO_LYRICS, value);
+		}
+
+		else if (id == "WEBLINK" ||
+			 id == "ARTIST URL")	 info.SetOtherInfo(INFO_WEB_ARTIST,	value);
+
+		else if (id == "PUBLISHER URL")	 info.SetOtherInfo(INFO_WEB_PUBLISHER,	value);
+		else if (id == "FILE URL")	 info.SetOtherInfo(INFO_WEB_SOURCE,	value);
+		else if (id == "COPYRIGHT URL")	 info.SetOtherInfo(INFO_WEB_COPYRIGHT,	value);
+		else if (id == "BUY URL")	 info.SetOtherInfo(INFO_WEB_COMMERCIAL,	value);
 
 		else if (id == "TRACK")
 		{
@@ -448,19 +511,11 @@ Error BoCA::TaggerAPEv2::ParseBuffer(const Buffer<UnsignedByte> &buffer, Track &
 		}
 		else if (id == "MCDI")
 		{
-			/* Check validity of MCDI data.
+			/* Check validity of MCDI data and assign.
 			 */
 			MCDI	 mcdi(item);
-			Bool	 valid = True;
 
-			for (Int i = 1; i < mcdi.GetNumberOfEntries(); i++)
-			{
-				if (mcdi.GetNthEntryOffset(i - 1) >= mcdi.GetNthEntryOffset(i)) valid = False;
-			}
-
-			/* Found a binary MCDI field.
-			 */
-			if (valid) info.mcdi.SetData(item);
+			if (mcdi.IsValid()) info.mcdi.SetData(item);
 		}
 		else if (id.StartsWith("COVER ART"))
 		{
@@ -506,7 +561,7 @@ Error BoCA::TaggerAPEv2::ParseBuffer(const Buffer<UnsignedByte> &buffer, Track &
 				/* Output cuesheet to temporary file.
 				 */
 				String		 cuesheet = value.Replace("\r\n", "\n");
-				String		 cueFile  = S::System::System::GetTempDirectory().Append("cuesheet_temp_").Append(String::FromInt(S::System::System::Clock())).Append(".cue");
+				String		 cueFile  = S::System::System::GetTempDirectory().Append("cuesheet_temp_").Append(String::FromInt(track.fileName.ComputeCRC32())).Append(".cue");
 				OutStream	 out(STREAM_FILE, cueFile, OS_REPLACE);
 
 				/* Write UTF-8 BOM.
@@ -536,6 +591,9 @@ Error BoCA::TaggerAPEv2::ParseBuffer(const Buffer<UnsignedByte> &buffer, Track &
 					Track	 cueTrack;
 					Config	*cueConfig = Config::Copy(GetConfiguration());
 
+					cueConfig->SetIntValue("Tags", "ReadChapters", False);
+					cueConfig->SetIntValue("Tags", "ReadEmbeddedCueSheets", False);
+
 					cueConfig->SetIntValue("CueSheet", "ReadInformationTags", True);
 					cueConfig->SetIntValue("CueSheet", "PreferCueSheets", True);
 					cueConfig->SetIntValue("CueSheet", "LookForAlternativeFiles", False);
@@ -566,6 +624,10 @@ Error BoCA::TaggerAPEv2::ParseBuffer(const Buffer<UnsignedByte> &buffer, Track &
 	 */
 	if (info.artist == NIL) info.artist = info.GetOtherInfo(INFO_ALBUMARTIST);
 
+	/* Remove sub-tracks if main track has a track number.
+	 */
+	if (info.track > 0) track.tracks.RemoveAll();
+
 	track.SetInfo(info);
 
 	return Success();
@@ -645,7 +707,7 @@ Bool BoCA::TaggerAPEv2::ParseAPEFooter(const Buffer<UnsignedByte> &buffer, Int *
 	return True;
 }
 
-Bool BoCA::TaggerAPEv2::ParseAPEItem(const Buffer<UnsignedByte> &buffer, Int &offset, String *id, String *value)
+Bool BoCA::TaggerAPEv2::ParseAPEItem(const Buffer<UnsignedByte> &buffer, Int &offset, String *id, String *value, Bool trim)
 {
 	InStream	 in(STREAM_BUFFER, buffer + offset, buffer.Size() - offset - 32);
 
@@ -673,7 +735,9 @@ Bool BoCA::TaggerAPEv2::ParseAPEItem(const Buffer<UnsignedByte> &buffer, Int &of
 	}
 	while (lastChar != 0);
 
-	*value = in.InputString(valueBytes).Trim();
+	*value = in.InputString(valueBytes);
+
+	if (trim) *value = value->Trim();
 
 	offset += in.GetPos();
 
