@@ -1,5 +1,5 @@
  /* BoCA - BonkEnc Component Architecture
-  * Copyright (C) 2007-2020 Robert Kausch <robert.kausch@freac.org>
+  * Copyright (C) 2007-2021 Robert Kausch <robert.kausch@freac.org>
   *
   * This program is free software; you can redistribute it and/or
   * modify it under the terms of the GNU General Public License as
@@ -80,6 +80,7 @@ const String	 BoCA::TaggerID3v2::genres[192] =
 
 BoCA::TaggerID3v2::TaggerID3v2()
 {
+	textEncoding = ID3TE_NONE;
 }
 
 BoCA::TaggerID3v2::~TaggerID3v2()
@@ -155,6 +156,19 @@ Error BoCA::TaggerID3v2::UpdateStreamInfo(const String &fileName, const Track &t
 {
 	InStream	 in(STREAM_FILE, fileName, IS_READ);
 
+	/* Check for RIFF or AIFF files.
+	 */
+	String	 magic = in.InputString(4);
+
+	if (magic == "RIFF" || magic == "FORM")
+	{
+		in.Close();
+
+		return UpdateChunkedFileTag(fileName, track);
+	}
+
+	in.Seek(0);
+
 	/* Look for ID3v2 tag.
 	 */
 	if (in.InputString(3) == "ID3" && in.InputNumber(1) <= 4)
@@ -175,6 +189,10 @@ Error BoCA::TaggerID3v2::UpdateStreamInfo(const String &fileName, const Track &t
 		 */
 		in.RelSeek(tagSize);
 	}
+	else
+	{
+		in.Seek(0);
+	}
 
 	/* Copy to temporary file and write tag.
 	 */
@@ -210,14 +228,145 @@ Error BoCA::TaggerID3v2::UpdateStreamInfo(const String &fileName, const Track &t
 	return Error();
 }
 
+Error BoCA::TaggerID3v2::UpdateChunkedFileTag(const String &fileName, const Track &track)
+{
+	InStream	 in(STREAM_FILE, fileName, IS_READ);
+
+	/* Read RIFF chunk.
+	 */
+	Bool		 error = False;
+	String		 riff  = in.InputString(4);
+
+	if (riff != "RIFF" && riff != "RF64" && riff != "FORM") return Error();
+
+	/* Create output file.
+	 */
+	OutStream	 out(STREAM_FILE, fileName.Append(".temp"), OS_REPLACE);
+
+	UnsignedInt32	 rSize	 = riff != "FORM" ? in.InputNumber(4) : in.InputNumberRaw(4);
+	UnsignedInt32	 rFormat = in.InputNumber(4);
+
+	out.OutputString(riff);
+
+	if (riff != "FORM") out.OutputNumber(rSize, 4);
+	else		    out.OutputNumberRaw(rSize, 4);
+
+	out.OutputNumber(rFormat, 4);
+
+	Int64			 dSize = -1;
+	Buffer<UnsignedByte>	 buffer(131072);
+
+	while (!error)
+	{
+		if (in.GetPos() >= in.Size()) break;
+
+		/* Read next chunk.
+		 */
+		String		 chunk = in.InputString(4);
+		UnsignedInt64	 cSize = (UnsignedInt32) (riff != "FORM" ? in.InputNumber(4) : in.InputNumberRaw(4));
+
+		if ((riff != "FORM" && chunk == "id3 ") ||
+		    (riff == "FORM" && chunk == "ID3 "))
+		{
+			/* Skip ID3 chunk.
+			 */
+			if (!in.RelSeek(cSize + cSize % 2)) error = True;
+
+			continue;
+		}
+
+		out.OutputString(chunk);
+
+		if (riff != "FORM") out.OutputNumber(cSize, 4);
+		else		    out.OutputNumberRaw(cSize, 4);
+
+		if (chunk == "ds64")
+		{
+			in.RelSeek(8);
+
+			dSize = in.InputNumber(8);
+
+			in.RelSeek(-16);
+		}
+		else if (chunk == "data")
+		{
+			if (rSize == 0xFFFFFFFF || rSize == 0 ||
+			    cSize == 0xFFFFFFFF || cSize == 0) error = True;
+
+			if (dSize >= 0) cSize = dSize;
+		}
+
+		/* Write chunk data to output file.
+		 */
+		cSize += cSize % 2;
+
+		while (!error && cSize > 0)
+		{
+			Int64	 bytes = Math::Min(cSize, buffer.Size());
+
+			in.InputData(buffer, bytes);
+			out.OutputData(buffer, bytes);
+
+			cSize -= bytes;
+		}
+	}
+
+	/* Write new metadata.
+	 */
+	if (!error)
+	{
+		/* Write ID3 chunk.
+		 */
+		buffer.Resize(0);
+
+		RenderBuffer(buffer, track);
+
+		if (riff != "FORM") out.OutputString("id3 ");
+		else		    out.OutputString("ID3 ");
+
+		Int	 size = buffer.Size();
+
+		if (riff != "FORM") out.OutputNumber(size, 4);
+		else		    out.OutputNumberRaw(size, 4);
+
+		out.OutputData(buffer, size);
+
+		/* Update file size.
+		 */
+		UnsignedInt64	 fileSize = out.GetPos() - 8;
+
+		rSize = Math::Min(0xFFFFFFFF, fileSize);
+
+		out.Seek(4);
+
+		if (riff != "FORM") out.OutputNumber(rSize, 4);
+		else		    out.OutputNumberRaw(rSize, 4);
+	}
+
+	/* Clean up.
+	 */
+	in.Close();
+	out.Close();
+
+	if (error)
+	{
+		File(fileName.Append(".temp")).Delete();
+
+		return Error();
+	}
+
+	File(fileName).Delete();
+	File(fileName.Append(".temp")).Move(fileName);
+
+	return Success();
+}
+
 Int BoCA::TaggerID3v2::RenderContainer(ID3_Container &container, const Track &track, Bool isChapter)
 {
 	/* Get configuration.
 	 */
 	const Config	*currentConfig		 = GetConfiguration();
 
-	String		 encodingID		 = currentConfig->GetStringValue(ConfigID, "ID3v2Encoding", "UTF-16LE");
-
 	Bool		 prependZero		 = currentConfig->GetIntValue(ConfigID, "TrackPrependZeroID3v2", False);
 
 	Bool		 writeChapters		 = currentConfig->GetIntValue(ConfigID, "WriteChapters", True);
@@ -231,10 +380,20 @@ Int BoCA::TaggerID3v2::RenderContainer(ID3_Container &container, const Track &tr
 	Bool		 replaceExistingComments = currentConfig->GetIntValue(ConfigID, "ReplaceExistingComments", False);
 	String		 defaultComment		 = currentConfig->GetStringValue(ConfigID, "DefaultComment", NIL);
 
+	/* Set text encoding.
+	 */
+	textEncodingID = currentConfig->GetStringValue(ConfigID, "ID3v2Encoding", "UTF-16LE");
+
+	if	(textEncodingID == "UTF-8")				      textEncoding = ID3TE_UTF8;
+	else if (textEncodingID == "ISO-8859-1")			      textEncoding = ID3TE_ISO8859_1;
+	else if (textEncodingID == "UTF-16"   || textEncodingID == "UCS-2" ||
+		 textEncodingID == "UTF-16LE" || textEncodingID == "UCS-2LE") textEncoding = ID3TE_UTF16;
+	else if (textEncodingID == "UTF-16BE" || textEncodingID == "UCS-2BE") textEncoding = ID3TE_UTF16BE;
+
 	/* Set ID3v2 version according to encoding.
 	 */
-	if (encodingID == "UTF-16BE" || encodingID == "UTF-8") container.SetSpec(ID3V2_4_0);
-	else						       container.SetSpec(ID3V2_3_0);
+	if (textEncodingID == "UTF-16BE" || textEncodingID == "UTF-8") container.SetSpec(ID3V2_4_0);
+	else							       container.SetSpec(ID3V2_3_0);
 
 	/* Save basic information.
 	 */
@@ -292,7 +451,7 @@ Int BoCA::TaggerID3v2::RenderContainer(ID3_Container &container, const Track &tr
 
 		SetASCIIField(frame, ID3FN_LANGUAGE, "eng"); // iTunes recognizes comments only if language is set to English
 
-		if (info.comment != NIL && !replaceExistingComments) SetStringField(frame, ID3FN_TEXT, info.comment);
+		if (info.comment != NIL && !replaceExistingComments) SetStringField(frame, ID3FN_TEXT, info.comment, False);
 		else						     SetStringField(frame, ID3FN_TEXT, defaultComment);
 
 		container.AddFrame(frame);
@@ -315,7 +474,7 @@ Int BoCA::TaggerID3v2::RenderContainer(ID3_Container &container, const Track &tr
 
 		if (value == NIL) continue;
 
-		if	(key == INFO_ALBUMARTIST)    { ID3_Frame frame(ID3FID_USERTEXT);	  SetStringField(frame, ID3FN_TEXT, value); SetStringField(frame, ID3FN_DESCRIPTION, "Album Artist"); container.AddFrame(frame); }
+		if	(key == INFO_ALBUMARTIST)    { ID3_Frame frame(ID3FID_USERTEXT);	  SetStringField(frame, ID3FN_TEXT, value); SetStringField(frame, ID3FN_DESCRIPTION, "Album Artist");  container.AddFrame(frame); }
 
 		else if	(key == INFO_CONTENTGROUP)   { ID3_Frame frame(ID3FID_CONTENTGROUP);	  SetStringField(frame, ID3FN_TEXT, value); container.AddFrame(frame); }
 		else if	(key == INFO_SUBTITLE)	     { ID3_Frame frame(ID3FID_SUBTITLE);	  SetStringField(frame, ID3FN_TEXT, value); container.AddFrame(frame); }
@@ -342,13 +501,28 @@ Int BoCA::TaggerID3v2::RenderContainer(ID3_Container &container, const Track &tr
 			container.AddFrame(frame);
 		}
 
+		else if	(key == INFO_MOVEMENT)
+		{
+			if (info.HasOtherInfo(INFO_MOVEMENTTOTAL)) value.Append("/").Append(info.GetOtherInfo(INFO_MOVEMENTTOTAL));
+
+			{ ID3_Frame frame(ID3FID_MOVEMENT); SetStringField(frame, ID3FN_TEXT, value); container.AddFrame(frame); }
+		}
+
+		else if	(key == INFO_MOVEMENTNAME)   { ID3_Frame frame(ID3FID_MOVEMENTNAME);	  SetStringField(frame, ID3FN_TEXT, value); container.AddFrame(frame); }
+
 		else if	(key == INFO_BPM)	     { ID3_Frame frame(ID3FID_BPM);		  SetStringField(frame, ID3FN_TEXT, value); container.AddFrame(frame); }
 		else if	(key == INFO_INITIALKEY)     { ID3_Frame frame(ID3FID_INITIALKEY);	  SetStringField(frame, ID3FN_TEXT, value); container.AddFrame(frame); }
 
 		else if	(key == INFO_COPYRIGHT)	     { ID3_Frame frame(ID3FID_COPYRIGHT);	  SetStringField(frame, ID3FN_TEXT, value); container.AddFrame(frame); }
 
+		else if	(key == INFO_MEDIATYPE)	     { ID3_Frame frame(ID3FID_MEDIATYPE);	  SetStringField(frame, ID3FN_TEXT, value); container.AddFrame(frame); }
+		else if (key == INFO_CATALOGNUMBER)  { ID3_Frame frame(ID3FID_USERTEXT);	  SetStringField(frame, ID3FN_TEXT, value); SetStringField(frame, ID3FN_DESCRIPTION, "CATALOGNUMBER"); container.AddFrame(frame); }
+		else if (key == INFO_BARCODE)	     { ID3_Frame frame(ID3FID_USERTEXT);	  SetStringField(frame, ID3FN_TEXT, value); SetStringField(frame, ID3FN_DESCRIPTION, "BARCODE");       container.AddFrame(frame); }
+
 		else if	(key == INFO_DISCSUBTITLE)   { ID3_Frame frame(ID3FID_SETSUBTITLE);	  SetStringField(frame, ID3FN_TEXT, value); container.AddFrame(frame); }
 
+		else if	(key == INFO_LYRICS)	     { ID3_Frame frame(ID3FID_UNSYNCEDLYRICS);	  SetStringField(frame, ID3FN_TEXT, value, False); SetASCIIField(frame, ID3FN_LANGUAGE, "und"); container.AddFrame(frame); }
+
 		else if	(key == INFO_RADIOSTATION)   { ID3_Frame frame(ID3FID_NETRADIOSTATION);   SetStringField(frame, ID3FN_TEXT, value); container.AddFrame(frame); }
 		else if	(key == INFO_RADIOOWNER)     { ID3_Frame frame(ID3FID_NETRADIOOWNER);     SetStringField(frame, ID3FN_TEXT, value); container.AddFrame(frame); }
 
@@ -361,7 +535,7 @@ Int BoCA::TaggerID3v2::RenderContainer(ID3_Container &container, const Track &tr
 		else if	(key == INFO_WEB_COPYRIGHT)  { ID3_Frame frame(ID3FID_WWWCOPYRIGHT);      SetASCIIField(frame, ID3FN_URL, value);   container.AddFrame(frame); }
 		else if	(key == INFO_WEB_COMMERCIAL) { ID3_Frame frame(ID3FID_WWWCOMMERCIALINFO); SetASCIIField(frame, ID3FN_URL, value);   container.AddFrame(frame); }
 
-		else if (key == INFO_WEB_USERURL)    { ID3_Frame frame(ID3FID_WWWUSER);		  SetASCIIField(frame, ID3FN_URL, value.Tail(value.Length() - value.Find(":|:") - 3)); SetStringField(frame, ID3FN_DESCRIPTION, value.Head(value.Find(":|:"))); container.AddFrame(frame); }
+		else if (key == INFO_WEB_USERURL)    { ID3_Frame frame(ID3FID_WWWUSER);		  SetASCIIField(frame, ID3FN_URL, value.Tail(value.Length() - value.Find(":|:") - 3));	 SetStringField(frame, ID3FN_DESCRIPTION, value.Head(value.Find(":|:"))); container.AddFrame(frame); }
 	}
 
 	/* Save Replay Gain info.
@@ -383,16 +557,13 @@ Int BoCA::TaggerID3v2::RenderContainer(ID3_Container &container, const Track &tr
 
 	/* Save CD table of contents.
 	 */
-	if (writeMCDI && !isChapter)
+	if (writeMCDI && info.mcdi.IsValid() && !isChapter)
 	{
-		if (info.mcdi.GetData().Size() > 0)
-		{
-			ID3_Frame	 frame_mcdi(ID3FID_CDID);
+		ID3_Frame	 frame_mcdi(ID3FID_CDID);
 
-			SetBinaryField(frame_mcdi, ID3FN_DATA, info.mcdi.GetData());
+		SetBinaryField(frame_mcdi, ID3FN_DATA, info.mcdi.GetData());
 
-			container.AddFrame(frame_mcdi);
-		}
+		container.AddFrame(frame_mcdi);
 	}
 
 	/* Save encoder version.
@@ -420,18 +591,19 @@ Int BoCA::TaggerID3v2::RenderContainer(ID3_Container &container, const Track &tr
 			 */
 			if (picInfo.description != NIL)
 			{
-				Config	*singleByteConfig = Config::Copy(currentConfig);
-				String	 encoding	  = encodingID;
+				String		 savedEncodingID = textEncodingID;
+				ID3_TextEnc	 savedEncoding	 = textEncoding;
 
-				if (encoding != "UTF-8" && !String::IsUnicode(picInfo.description)) singleByteConfig->SetStringValue(ConfigID, "ID3v2Encoding", "ISO-8859-1");
-
-				SetConfiguration(singleByteConfig);
+				if (textEncodingID != "UTF-8" && !String::IsUnicode(picInfo.description))
+				{
+					textEncodingID = "ISO-8859-1";
+					textEncoding   = ID3TE_ISO8859_1;
+				}
 
 				SetStringField(frame_picture, ID3FN_DESCRIPTION, picInfo.description);
 
-				SetConfiguration(currentConfig);
-
-				Config::Free(singleByteConfig);
+				textEncodingID = savedEncodingID;
+				textEncoding   = savedEncoding;
 			}
 
 			/* Set picture data.
@@ -505,6 +677,7 @@ Int BoCA::TaggerID3v2::ParseContainer(const ID3_Container &container, Track &tra
 	 */
 	ID3_Container::ConstIterator	*iterator     = container.CreateIterator();
 	Bool				 haveChapters = False;
+	String				 overDriveMarkers;
 
 	for (UnsignedInt i = 0; i < container.NumFrames(); i++)
 	{
@@ -515,9 +688,13 @@ Int BoCA::TaggerID3v2::ParseContainer(const ID3_Container &container, Track &tra
 		else if (frame.GetID() == ID3FID_ALBUM)		    info.album	 = GetStringField(frame, ID3FN_TEXT);
 		else if (frame.GetID() == ID3FID_YEAR)		    info.year	 = GetStringField(frame, ID3FN_TEXT).ToInt();
 		else if (frame.GetID() == ID3FID_RECORDINGTIME)	    info.year	 = GetStringField(frame, ID3FN_TEXT).Head(4).ToInt();
-		else if (frame.GetID() == ID3FID_COMMENT)	    info.comment = GetStringField(frame, ID3FN_TEXT);
+		else if (frame.GetID() == ID3FID_COMMENT)	    info.comment = GetStringField(frame, ID3FN_TEXT, False);
 		else if (frame.GetID() == ID3FID_PUBLISHER)	    info.label	 = GetStringField(frame, ID3FN_TEXT);
-		else if (frame.GetID() == ID3FID_ISRC)		    info.isrc	 = GetStringField(frame, ID3FN_TEXT);
+
+		else if (frame.GetID() == ID3FID_ISRC)
+		{
+			if (Info::IsISRC(GetStringField(frame, ID3FN_TEXT))) info.isrc = GetStringField(frame, ID3FN_TEXT);
+		}
 
 		else if (frame.GetID() == ID3FID_CONTENTGROUP)	    info.SetOtherInfo(INFO_CONTENTGROUP,   GetStringField(frame, ID3FN_TEXT));
 		else if (frame.GetID() == ID3FID_SUBTITLE)	    info.SetOtherInfo(INFO_SUBTITLE,	   GetStringField(frame, ID3FN_TEXT));
@@ -534,13 +711,32 @@ Int BoCA::TaggerID3v2::ParseContainer(const ID3_Container &container, Track &tra
 		else if (frame.GetID() == ID3FID_ORIGYEAR)	    info.SetOtherInfo(INFO_ORIG_YEAR,	   GetStringField(frame, ID3FN_TEXT));
 		else if (frame.GetID() == ID3FID_ORIGRELEASETIME)   info.SetOtherInfo(INFO_ORIG_YEAR,	   GetStringField(frame, ID3FN_TEXT).Head(4));
 
-		else if (frame.GetID() == ID3FID_BPM)		    info.SetOtherInfo(INFO_BPM,		   GetStringField(frame, ID3FN_TEXT));
+		else if (frame.GetID() == ID3FID_MOVEMENT)
+		{
+			String	 movementString = GetStringField(frame, ID3FN_TEXT);
+
+			info.SetOtherInfo(INFO_MOVEMENT, String::FromInt(movementString.ToInt()));
+
+			if (movementString.Contains("/")) info.SetOtherInfo(INFO_MOVEMENTTOTAL, movementString.Tail(movementString.Length() - movementString.Find("/") - 1));
+		}
+
+		else if (frame.GetID() == ID3FID_MOVEMENTNAME)	    info.SetOtherInfo(INFO_MOVEMENTNAME,   GetStringField(frame, ID3FN_TEXT));
+
+		else if (frame.GetID() == ID3FID_BPM)
+		{
+			if (GetStringField(frame, ID3FN_TEXT).ToInt() > 0) info.SetOtherInfo(INFO_BPM, GetStringField(frame, ID3FN_TEXT));
+		}
+
 		else if (frame.GetID() == ID3FID_INITIALKEY)	    info.SetOtherInfo(INFO_INITIALKEY,	   GetStringField(frame, ID3FN_TEXT));
 
 		else if (frame.GetID() == ID3FID_COPYRIGHT)	    info.SetOtherInfo(INFO_COPYRIGHT,	   GetStringField(frame, ID3FN_TEXT));
 
+		else if (frame.GetID() == ID3FID_MEDIATYPE)	    info.SetOtherInfo(INFO_MEDIATYPE,	   GetStringField(frame, ID3FN_TEXT));
+
 		else if (frame.GetID() == ID3FID_SETSUBTITLE)	    info.SetOtherInfo(INFO_DISCSUBTITLE,   GetStringField(frame, ID3FN_TEXT));
 
+		else if (frame.GetID() == ID3FID_UNSYNCEDLYRICS)    info.SetOtherInfo(INFO_LYRICS,	   GetStringField(frame, ID3FN_TEXT, False));
+
 		else if (frame.GetID() == ID3FID_NETRADIOSTATION)   info.SetOtherInfo(INFO_RADIOSTATION,   GetStringField(frame, ID3FN_TEXT));
 		else if (frame.GetID() == ID3FID_NETRADIOOWNER)     info.SetOtherInfo(INFO_RADIOOWNER,	   GetStringField(frame, ID3FN_TEXT));
 
@@ -603,7 +799,7 @@ Int BoCA::TaggerID3v2::ParseContainer(const ID3_Container &container, Track &tra
  					/* Output cuesheet to temporary file.
 					 */
 					String		 cuesheet = value.Replace("\r\n", "\n");
-					String		 cueFile  = S::System::System::GetTempDirectory().Append("cuesheet_temp_").Append(String::FromInt(S::System::System::Clock())).Append(".cue");
+					String		 cueFile  = S::System::System::GetTempDirectory().Append("cuesheet_temp_").Append(String::FromInt(track.fileName.ComputeCRC32())).Append(".cue");
 					OutStream	 out(STREAM_FILE, cueFile, OS_REPLACE);
 
 					/* Write UTF-8 BOM.
@@ -633,6 +829,9 @@ Int BoCA::TaggerID3v2::ParseContainer(const ID3_Container &container, Track &tra
 						Track	 cueTrack;
 						Config	*cueConfig = Config::Copy(GetConfiguration());
 
+						cueConfig->SetIntValue("Tags", "ReadChapters", False);
+						cueConfig->SetIntValue("Tags", "ReadEmbeddedCueSheets", False);
+
 						cueConfig->SetIntValue("CueSheet", "ReadInformationTags", True);
 						cueConfig->SetIntValue("CueSheet", "PreferCueSheets", True);
 						cueConfig->SetIntValue("CueSheet", "LookForAlternativeFiles", False);
@@ -652,7 +851,17 @@ Int BoCA::TaggerID3v2::ParseContainer(const ID3_Container &container, Track &tra
 				}
 			}
 
-			else if (description.ToLower() == "album artist")	   info.SetOtherInfo(INFO_ALBUMARTIST, value);
+			else if (description.ToLower() == "album artist")	   info.SetOtherInfo(INFO_ALBUMARTIST,	 value);
+
+			else if (description.ToLower() == "catalognumber")	   info.SetOtherInfo(INFO_CATALOGNUMBER, value);
+			else if (description.ToLower() == "barcode")		   info.SetOtherInfo(INFO_BARCODE,	 value);
+
+			else if (description == "OverDrive MediaMarkers")
+			{
+				/* Save OverDrive markers to process as chapters further down.
+				 */
+				overDriveMarkers = value;
+			}
 
 			else							   info.other.Add(String(INFO_USERTEXT).Append(":").Append(description).Append(":|:").Append(value));
 		}
@@ -693,19 +902,11 @@ Int BoCA::TaggerID3v2::ParseContainer(const ID3_Container &container, Track &tra
 
 			if (binary)
 			{
-				/* Check validity of MCDI data.
+				/* Check validity of MCDI data and assign.
 				 */
 				MCDI	 mcdi(data);
-				Bool	 valid = True;
 
-				for (Int i = 1; i < mcdi.GetNumberOfEntries(); i++)
-				{
-					if (mcdi.GetNthEntryOffset(i - 1) >= mcdi.GetNthEntryOffset(i)) valid = False;
-				}
-
-				/* Found a binary MCDI field.
-				 */
-				if (valid) info.mcdi.SetData(data);
+				if (mcdi.IsValid()) info.mcdi.SetData(data);
 			}
 			else
 			{
@@ -763,6 +964,10 @@ Int BoCA::TaggerID3v2::ParseContainer(const ID3_Container &container, Track &tra
 	 */
 	if (info.artist == NIL) info.artist = info.GetOtherInfo(INFO_ALBUMARTIST);
 
+	/* Remove sub-tracks if main track has a track number.
+	 */
+	if (info.track > 0) track.tracks.RemoveAll();
+
 	track.SetInfo(info);
 
 	delete iterator;
@@ -863,10 +1068,95 @@ Int BoCA::TaggerID3v2::ParseContainer(const ID3_Container &container, Track &tra
 		}
 	}
 
+	/* Read OverDrive MediaMarkers.
+	 */
+	if (!haveChapters && overDriveMarkers != NIL && currentConfig->GetIntValue(ConfigID, "ReadChapters", True) &&
+						      (!currentConfig->GetIntValue(ConfigID, "PreferCueSheetsToChapters", True) || track.tracks.Length() == 0))
+	{
+		track.tracks.RemoveAll();
+
+		if (!overDriveMarkers.StartsWith("<?xml")) overDriveMarkers = String("<?xml version=\"1.0\" encoding=\"UTF-8\"?>").Append(overDriveMarkers);
+
+		XML::Document	 doc;
+		const char	*xml = overDriveMarkers.ConvertTo("UTF-8");
+
+		if (doc.ParseMemory(xml, strlen(xml)) == Success())
+		{
+			XML::Node	*root	      = doc.GetRootNode();
+			const Format	&format	      = track.GetFormat();
+			Int		 chapterCount = 0;
+
+			for (Int i = 0; i < root->GetNOfNodes(); i++)
+			{
+				XML::Node	*marker = root->GetNthNode(i);
+
+				if (marker->GetName() != "Marker") continue;
+
+				String	 chapterName;
+				String	 chapterTime;
+
+				if (XML::Node *name = marker->GetNodeByName("Name")) chapterName = name->GetContent().Trim();
+				if (XML::Node *time = marker->GetNodeByName("Time")) chapterTime = time->GetContent().Trim();
+
+				chapterCount++;
+
+				/* Fill track data.
+				 */
+				Track	 rTrack;
+
+				rTrack.fileName	    = track.fileName;
+				rTrack.pictures	    = track.pictures;
+
+				Array<String>	 chapterTimes = chapterTime.Explode(":");
+
+				if (chapterTimes.Length() == 2) chapterTimes.InsertAtPos(0, "00");
+
+				rTrack.sampleOffset  =		   chapterTimes.GetNth(0).ToInt() * 60 * 60 * format.rate;
+				rTrack.sampleOffset +=		   chapterTimes.GetNth(1).ToInt() * 60      * format.rate;
+				rTrack.sampleOffset += Math::Round(chapterTimes.GetNth(2).ToFloat()	    * format.rate);
+
+				if	(track.length	    > 0) rTrack.length	     = track.length	  - rTrack.sampleOffset;
+				else if (track.approxLength > 0) rTrack.approxLength = track.approxLength - rTrack.sampleOffset;
+
+				if	(track.length	    > 0) rTrack.fileSize = Math::Round(Float(track.fileSize) / track.length * rTrack.length);
+				else if (track.approxLength > 0) rTrack.fileSize = Math::Round(Float(track.fileSize) / track.approxLength * rTrack.length);
+				else				 rTrack.fileSize = rTrack.length * format.channels * (format.bits / 8);
+
+				rTrack.SetFormat(format);
+
+				/* Set track title.
+				 */
+				Info	 info = track.GetInfo();
+
+				info.title = chapterName;
+				info.track = chapterCount;
+
+				rTrack.SetInfo(info);
+
+				/* Update previous track length.
+				 */
+				if (chapterCount > 1)
+				{
+					Track	&pTrack = track.tracks.GetReference(chapterCount - 1);
+
+					pTrack.length	= rTrack.sampleOffset - pTrack.sampleOffset;
+
+					if	(track.length	    > 0) pTrack.fileSize = Math::Round(Float(track.fileSize) / track.length * pTrack.length);
+					else if (track.approxLength > 0) pTrack.fileSize = Math::Round(Float(track.fileSize) / track.approxLength * pTrack.length);
+					else				 pTrack.fileSize = pTrack.length * format.channels * (format.bits / 8);
+				}
+
+				/* Add track to track list.
+				 */
+				track.tracks.Add(rTrack);
+			}
+		}
+	}
+
 	return Success();
 }
 
-String BoCA::TaggerID3v2::GetStringField(const ID3_Frame &frame, ID3_FieldID fieldType)
+String BoCA::TaggerID3v2::GetStringField(const ID3_Frame &frame, ID3_FieldID fieldType, Bool trim)
 {
 	ID3_Field	*field = frame.GetField(fieldType);
 	String		 result;
@@ -877,58 +1167,47 @@ String BoCA::TaggerID3v2::GetStringField(const ID3_Frame &frame, ID3_FieldID fie
 
 		if (encoding == ID3TE_ISO8859_1 || encoding == ID3TE_UTF8)
 		{
-			Buffer<char>	 aBuffer(1024);
+			Buffer<char>	 aBuffer(field->Size() + 1);
 
 			aBuffer.Zero();
 
-			field->Get(aBuffer, aBuffer.Size());
+			field->Get(aBuffer, field->Size());
 
 			if	(encoding == ID3TE_ISO8859_1) result.ImportFrom("ISO-8859-1", aBuffer);
 			else if (encoding == ID3TE_UTF8)      result.ImportFrom("UTF-8", aBuffer);
 		}
 		else if (encoding == ID3TE_UTF16 || encoding == ID3TE_UTF16BE)
 		{
-			Buffer<wchar_t>	 wBuffer(1024);
+			Buffer<wchar_t>	 wBuffer(field->Size() + 1);
 
 			wBuffer.Zero();
 
-			field->Get((unicode_t *) (wchar_t *) wBuffer, wBuffer.Size());
+			field->Get((unicode_t *) (wchar_t *) wBuffer, field->Size());
 
 			result.ImportFrom("UTF-16BE", (char *) (wchar_t *) wBuffer);
 		}
 	}
 
-	return result.Trim();
+	return trim ? result.Trim() : result;
 }
 
-Int BoCA::TaggerID3v2::SetStringField(ID3_Frame &frame, ID3_FieldID fieldType, const String &string)
+Int BoCA::TaggerID3v2::SetStringField(ID3_Frame &frame, ID3_FieldID fieldType, const String &value, Bool trim)
 {
+	String	 string = trim ? value.Trim() : value;
+
 	if (string == NIL) return Error();
 
-	const Config	*config = GetConfiguration();
-
-	ID3_TextEnc	 encoding   = ID3TE_NONE;
-	String		 encodingID = config->GetStringValue(ConfigID, "ID3v2Encoding", "UTF-16LE");
-
-	if	(encodingID == "UTF-8")				      encoding = ID3TE_UTF8;
-	else if (encodingID == "ISO-8859-1")			      encoding = ID3TE_ISO8859_1;
-	else if (encodingID == "UTF-16"	  || encodingID == "UCS-2" ||
-		 encodingID == "UTF-16LE" || encodingID == "UCS-2LE") encoding = ID3TE_UTF16;
-	else if (encodingID == "UTF-16BE" || encodingID == "UCS-2BE") encoding = ID3TE_UTF16BE;
-
-	String::OutputFormat	 outputFormat(encodingID);
-
-	SetIntegerField(frame, ID3FN_TEXTENC, encoding);
+	SetIntegerField(frame, ID3FN_TEXTENC, textEncoding);
 
 	ID3_Field	*field = frame.GetField(fieldType);
 
 	if (field != NIL)
 	{
-		field->SetEncoding(encoding);
+		field->SetEncoding(textEncoding);
 
-		if	(encoding == ID3TE_UTF16)   field->Set((unicode_t *) string.Trim().ConvertTo("UTF-16LE"));
-		else if (encoding == ID3TE_UTF16BE) field->Set((unicode_t *) string.Trim().ConvertTo("UTF-16BE"));
-		else				    field->Set((char *) string.Trim());
+		if	(textEncoding == ID3TE_UTF16)	field->Set((unicode_t *) string.ConvertTo("UTF-16LE"));
+		else if (textEncoding == ID3TE_UTF16BE)	field->Set((unicode_t *) string.ConvertTo("UTF-16BE"));
+		else					field->Set((char *)	 string.ConvertTo(textEncodingID));
 
 		return Success();
 	}
@@ -936,34 +1215,36 @@ Int BoCA::TaggerID3v2::SetStringField(ID3_Frame &frame, ID3_FieldID fieldType, c
 	return Error();
 }
 
-String BoCA::TaggerID3v2::GetASCIIField(const ID3_Frame &frame, ID3_FieldID fieldType)
+String BoCA::TaggerID3v2::GetASCIIField(const ID3_Frame &frame, ID3_FieldID fieldType, Bool trim)
 {
 	ID3_Field	*field = frame.GetField(fieldType);
 	String		 result;
 
 	if (field != NIL)
 	{
-		Buffer<char>	 aBuffer(1024);
+		Buffer<char>	 aBuffer(field->Size() + 1);
 
 		aBuffer.Zero();
 
-		field->Get(aBuffer, aBuffer.Size());
+		field->Get(aBuffer, field->Size());
 
 		result.ImportFrom("ISO-8859-1", aBuffer);
 	}
 
-	return result.Trim();
+	return trim ? result.Trim() : result;
 }
 
-Int BoCA::TaggerID3v2::SetASCIIField(ID3_Frame &frame, ID3_FieldID fieldType, const String &string)
+Int BoCA::TaggerID3v2::SetASCIIField(ID3_Frame &frame, ID3_FieldID fieldType, const String &value, Bool trim)
 {
+	String	 string = trim ? value.Trim() : value;
+
 	if (string == NIL) return Error();
 
 	ID3_Field	*field = frame.GetField(fieldType);
 
 	if (field != NIL)
 	{
-		field->Set((char *) string.Trim());
+		field->Set((char *) string);
 
 		return Success();
 	}
