@@ -1,5 +1,5 @@
  /* BoCA - BonkEnc Component Architecture
-  * Copyright (C) 2007-2020 Robert Kausch <robert.kausch@freac.org>
+  * Copyright (C) 2007-2021 Robert Kausch <robert.kausch@freac.org>
   *
   * This program is free software; you can redistribute it and/or
   * modify it under the terms of the GNU General Public License as
@@ -79,10 +79,14 @@ Void smooth::DetachDLL()
 	FreeWMVCoreDLL();
 }
 
-namespace BoCA
+static String toUNCPath(const String &streamURI)
 {
+	if (streamURI.StartsWith("\\\\")) return streamURI;
+
 	static const String	 uncPrefix = "\\\\?\\";
-};
+
+	return uncPrefix.Append(File(streamURI));
+}
 
 Void BoCA::TaggerWMA::Initialize()
 {
@@ -119,6 +123,8 @@ Error BoCA::TaggerWMA::RenderStreamInfo(const String &fileName, const Track &tra
 	Bool		 writeChapters		 = currentConfig->GetIntValue(ConfigID, "WriteChapters", True);
 	Bool		 writeMCDI		 = currentConfig->GetIntValue(ConfigID, "WriteMCDI", True);
 
+	Bool		 preserveReplayGain	 = currentConfig->GetIntValue(ConfigID, "PreserveReplayGain", True);
+
 	Bool		 coverArtWriteToTags	 = currentConfig->GetIntValue(ConfigID, "CoverArtWriteToTags", True);
 	Bool		 coverArtWriteToWMA	 = currentConfig->GetIntValue(ConfigID, "CoverArtWriteToWMAMetadata", True);
 
@@ -145,9 +151,7 @@ Error BoCA::TaggerWMA::RenderStreamInfo(const String &fileName, const Track &tra
 
 	/* Open file and render tags.
 	 */
-	String	 uncPath = String(fileName.StartsWith("\\\\") ? "" : uncPrefix).Append(fileName);
-
-	hr = metadataEditor2->OpenEx(uncPath, GENERIC_READ | GENERIC_WRITE, 0);
+	hr = metadataEditor2->OpenEx(toUNCPath(fileName), GENERIC_READ | GENERIC_WRITE, 0);
 
 	if (!FAILED(hr))
 	{
@@ -186,7 +190,7 @@ Error BoCA::TaggerWMA::RenderStreamInfo(const String &fileName, const Track &tra
 			RenderWMAIntegerItem(g_wszWMSharedUserRating, Math::Min(99, info.rating), pHeaderInfo);
 		}
 
-		if	(info.comment != NIL && !replaceExistingComments) RenderWMAStringItem(g_wszWMDescription, info.comment, pHeaderInfo);
+		if	(info.comment != NIL && !replaceExistingComments) RenderWMAStringItem(g_wszWMDescription, info.comment, pHeaderInfo, False);
 		else if (defaultComment != NIL)				  RenderWMAStringItem(g_wszWMDescription, defaultComment, pHeaderInfo);
 
 		/* Save other text info.
@@ -219,6 +223,12 @@ Error BoCA::TaggerWMA::RenderStreamInfo(const String &fileName, const Track &tra
 
 			else if	(key == INFO_COPYRIGHT)	     RenderWMAStringItem(g_wszWMCopyright,		 value, pHeaderInfo);
 
+			else if	(key == INFO_MEDIATYPE)	     RenderWMAStringItem("WM/Media",			 value, pHeaderInfo);
+			else if	(key == INFO_CATALOGNUMBER)  RenderWMAStringItem("WM/CatalogNo",		 value, pHeaderInfo);
+			else if	(key == INFO_BARCODE)	     RenderWMAStringItem("WM/Barcode",			 value, pHeaderInfo);
+
+			else if	(key == INFO_LYRICS)	     RenderWMAStringItem(g_wszWMLyrics,			 value, pHeaderInfo, False);
+
 			else if	(key == INFO_RADIOSTATION)   RenderWMAStringItem(g_wszWMRadioStationName,	 value, pHeaderInfo);
 			else if	(key == INFO_RADIOOWNER)     RenderWMAStringItem(g_wszWMRadioStationOwner,	 value, pHeaderInfo);
 
@@ -228,16 +238,27 @@ Error BoCA::TaggerWMA::RenderStreamInfo(const String &fileName, const Track &tra
 			else if	(key == INFO_WEB_COMMERCIAL) RenderWMAStringItem(g_wszWMPromotionURL,		 value, pHeaderInfo);
 		}
 
-		/* Save CD table of contents.
+		/* Save Replay Gain info.
 		 */
-		if (writeMCDI)
+		if (preserveReplayGain)
 		{
-			if (info.mcdi.GetData().Size() > 0)
+			if (info.track_gain != NIL && info.track_peak != NIL)
 			{
-				RenderWMABinaryItem(g_wszWMMCDI, info.mcdi.GetData(), pHeaderInfo);
+				RenderWMAStringItem("replaygain_track_gain", info.track_gain, pHeaderInfo);
+				RenderWMAStringItem("replaygain_track_peak", info.track_peak, pHeaderInfo);
+			}
+
+			if (info.album_gain != NIL && info.album_peak != NIL)
+			{
+				RenderWMAStringItem("replaygain_album_gain", info.album_gain, pHeaderInfo);
+				RenderWMAStringItem("replaygain_album_peak", info.album_peak, pHeaderInfo);
 			}
 		}
 
+		/* Save CD table of contents.
+		 */
+		if (writeMCDI && info.mcdi.IsValid()) RenderWMABinaryItem(g_wszWMMCDI, info.mcdi.GetData(), pHeaderInfo);
+
 		/* Save encoder version.
 		 */
 		Application	*app = Application::Get();
@@ -304,9 +325,9 @@ Error BoCA::TaggerWMA::RenderStreamInfo(const String &fileName, const Track &tra
 	else		return Success();
 }
 
-Error BoCA::TaggerWMA::RenderWMAStringItem(const String &id, const String &value, Void *headerInfo)
+Error BoCA::TaggerWMA::RenderWMAStringItem(const String &id, const String &value, Void *headerInfo, Bool trim)
 {
-	String	 data	  = value.Trim();
+	String	 data	  = trim ? value.Trim() : value;
 	Int	 dataSize = data != NIL ? wcslen(data) + 1 : 0;
 
 	if (data == NIL) return Success();
@@ -357,9 +378,7 @@ Error BoCA::TaggerWMA::ParseStreamInfo(const String &fileName, Track &track)
 
 	/* Open file and parse tags.
 	 */
-	String	 uncPath = String(fileName.StartsWith("\\\\") ? "" : uncPrefix).Append(fileName);
-
-	hr = metadataEditor2->OpenEx(uncPath, GENERIC_READ, FILE_SHARE_READ);
+	hr = metadataEditor2->OpenEx(toUNCPath(fileName), GENERIC_READ, FILE_SHARE_READ);
 
 	if (!FAILED(hr))
 	{
@@ -395,6 +414,7 @@ Error BoCA::TaggerWMA::ParseStreamInfo(const String &fileName, Track &track)
 
 			/* Parse string values.
 			 */
+			String	 id = name;
 			String	 value;
 
 			if (type == WMT_TYPE_STRING)
@@ -412,50 +432,64 @@ Error BoCA::TaggerWMA::ParseStreamInfo(const String &fileName, Track &track)
 
 			/* Assign values to metadata fields.
 			 */
-			if	(String(name) == g_wszWMAuthor)			 info.artist  = value;
-			else if (String(name) == g_wszWMTitle)			 info.title   = value;
-			else if (String(name) == g_wszWMAlbumTitle)		 info.album   = value;
-			else if (String(name) == g_wszWMYear)			 info.year    = value.ToInt();
-			else if (String(name) == g_wszWMGenre)			 info.genre   = value;
-			else if (String(name) == g_wszWMDescription)		 info.comment = value;
-			else if (String(name) == g_wszWMPublisher)		 info.label   = value;
-			else if (String(name) == g_wszWMISRC)			 info.isrc    = value;
+			if	(id == g_wszWMAuthor)		       info.artist  = value;
+			else if (id == g_wszWMTitle)		       info.title   = value;
+			else if (id == g_wszWMAlbumTitle)	       info.album   = value;
+			else if (id == g_wszWMYear)		       info.year    = value.ToInt();
+			else if (id == g_wszWMGenre)		       info.genre   = value;
+			else if (id == g_wszWMDescription)	       info.comment = String((LPWSTR) pbValue);
+			else if (id == g_wszWMPublisher)	       info.label   = value;
 
-			else if (String(name) == g_wszWMAlbumArtist)		 info.SetOtherInfo(INFO_ALBUMARTIST,	value);
+			else if (id == g_wszWMISRC)
+			{
+				if (Info::IsISRC(value)) info.isrc = value;
+			}
 
-			else if (String(name) == g_wszWMContentGroupDescription) info.SetOtherInfo(INFO_CONTENTGROUP,	value);
-			else if (String(name) == g_wszWMSubTitle)		 info.SetOtherInfo(INFO_SUBTITLE,	value);
+			else if (id == g_wszWMAlbumArtist)	       info.SetOtherInfo(INFO_ALBUMARTIST,    value);
 
-			else if (String(name) == g_wszWMConductor)		 info.SetOtherInfo(INFO_CONDUCTOR,	value);
-			else if (String(name) == g_wszWMModifiedBy)		 info.SetOtherInfo(INFO_REMIXER,	value);
-			else if (String(name) == g_wszWMComposer)		 info.SetOtherInfo(INFO_COMPOSER,	value);
-			else if (String(name) == g_wszWMWriter)			 info.SetOtherInfo(INFO_LYRICIST,	value);
-			else if (String(name) == g_wszWMProducer)		 info.SetOtherInfo(INFO_PRODUCER,	value);
+			else if (id == g_wszWMContentGroupDescription) info.SetOtherInfo(INFO_CONTENTGROUP,   value);
+			else if (id == g_wszWMSubTitle)		       info.SetOtherInfo(INFO_SUBTITLE,	      value);
 
-			else if (String(name) == g_wszWMOriginalArtist)		 info.SetOtherInfo(INFO_ORIG_ARTIST,	value);
-			else if (String(name) == g_wszWMOriginalAlbumTitle)	 info.SetOtherInfo(INFO_ORIG_ALBUM,	value);
-			else if (String(name) == g_wszWMOriginalLyricist)	 info.SetOtherInfo(INFO_ORIG_LYRICIST,	value);
-			else if (String(name) == g_wszWMOriginalReleaseYear)	 info.SetOtherInfo(INFO_ORIG_YEAR,	value);
+			else if (id == g_wszWMConductor)	       info.SetOtherInfo(INFO_CONDUCTOR,      value);
+			else if (id == g_wszWMModifiedBy)	       info.SetOtherInfo(INFO_REMIXER,	      value);
+			else if (id == g_wszWMComposer)		       info.SetOtherInfo(INFO_COMPOSER,	      value);
+			else if (id == g_wszWMWriter)		       info.SetOtherInfo(INFO_LYRICIST,	      value);
+			else if (id == g_wszWMProducer)		       info.SetOtherInfo(INFO_PRODUCER,	      value);
 
-			else if (String(name) == g_wszWMBeatsPerMinute)		 info.SetOtherInfo(INFO_BPM,		value);
-			else if (String(name) == g_wszWMInitialKey)		 info.SetOtherInfo(INFO_INITIALKEY,	value);
+			else if (id == g_wszWMOriginalArtist)	       info.SetOtherInfo(INFO_ORIG_ARTIST,    value);
+			else if (id == g_wszWMOriginalAlbumTitle)      info.SetOtherInfo(INFO_ORIG_ALBUM,     value);
+			else if (id == g_wszWMOriginalLyricist)	       info.SetOtherInfo(INFO_ORIG_LYRICIST,  value);
+			else if (id == g_wszWMOriginalReleaseYear)     info.SetOtherInfo(INFO_ORIG_YEAR,      value);
 
-			else if (String(name) == g_wszWMCopyright)		 info.SetOtherInfo(INFO_COPYRIGHT,	value);
+			else if (id == g_wszWMBeatsPerMinute)
+			{
+				if (value.ToInt() > 0) info.SetOtherInfo(INFO_BPM, value);
+			}
 
-			else if (String(name) == g_wszWMRadioStationName)	 info.SetOtherInfo(INFO_RADIOSTATION,	value);
-			else if (String(name) == g_wszWMRadioStationOwner)	 info.SetOtherInfo(INFO_RADIOOWNER,	value);
+			else if (id == g_wszWMInitialKey)	       info.SetOtherInfo(INFO_INITIALKEY,     value);
 
-			else if (String(name) == g_wszWMAuthorURL)		 info.SetOtherInfo(INFO_WEB_ARTIST,	value);
-			else if (String(name) == g_wszWMAudioSourceURL)		 info.SetOtherInfo(INFO_WEB_SOURCE,	value);
-			else if (String(name) == g_wszWMCopyrightURL)		 info.SetOtherInfo(INFO_WEB_COPYRIGHT,	value);
-			else if (String(name) == g_wszWMPromotionURL)		 info.SetOtherInfo(INFO_WEB_COMMERCIAL,	value);
+			else if (id == g_wszWMCopyright)	       info.SetOtherInfo(INFO_COPYRIGHT,      value);
 
-			else if (String(name) == g_wszWMTrack)
+			else if (id == "WM/Media")		       info.SetOtherInfo(INFO_MEDIATYPE,      value);
+			else if (id == "WM/CatalogNo")		       info.SetOtherInfo(INFO_CATALOGNUMBER,  value);
+			else if (id == "WM/Barcode")		       info.SetOtherInfo(INFO_BARCODE,	      value);
+
+			else if (id == g_wszWMLyrics)		       info.SetOtherInfo(INFO_LYRICS,	      String((LPWSTR) pbValue));
+
+			else if (id == g_wszWMRadioStationName)	       info.SetOtherInfo(INFO_RADIOSTATION,   value);
+			else if (id == g_wszWMRadioStationOwner)       info.SetOtherInfo(INFO_RADIOOWNER,     value);
+
+			else if (id == g_wszWMAuthorURL)	       info.SetOtherInfo(INFO_WEB_ARTIST,     value);
+			else if (id == g_wszWMAudioSourceURL)	       info.SetOtherInfo(INFO_WEB_SOURCE,     value);
+			else if (id == g_wszWMCopyrightURL)	       info.SetOtherInfo(INFO_WEB_COPYRIGHT,  value);
+			else if (id == g_wszWMPromotionURL)	       info.SetOtherInfo(INFO_WEB_COMMERCIAL, value);
+
+			else if (id == g_wszWMTrack)
 			{
 				if	(type == WMT_TYPE_DWORD)  info.track = 1 + ((DWORD *) pbValue)[0];
 				else if (type == WMT_TYPE_STRING) info.track = 1 + value.ToInt();
 			}
-			else if (String(name) == g_wszWMTrackNumber)
+			else if (id == g_wszWMTrackNumber)
 			{
 				if	(type == WMT_TYPE_DWORD)  info.track = ((DWORD *) pbValue)[0];
 				else if (type == WMT_TYPE_STRING)
@@ -465,19 +499,26 @@ Error BoCA::TaggerWMA::ParseStreamInfo(const String &fileName, Track &track)
 					if (value.Contains("/")) info.numTracks = value.Tail(value.Length() - value.Find("/") - 1).ToInt();
 				}
 			}
-			else if (String(name) == g_wszWMPartOfSet)
+			else if (id == g_wszWMPartOfSet)
 			{
 				info.disc = value.ToInt();
 
 				if (value.Contains("/")) info.numDiscs = value.Tail(value.Length() - value.Find("/") - 1).ToInt();
 			}
-			else if (String(name) == g_wszWMSharedUserRating)
+			else if (id == g_wszWMSharedUserRating)
 			{
 				info.rating = (DWORD) *pbValue;
 
 				if (info.rating == 99) info.rating = 100;
 			}
-			else if (String(name) == g_wszWMMCDI)
+			else if (id.ToUpper().StartsWith("REPLAYGAIN"))
+			{
+				if	(id.ToUpper() == "REPLAYGAIN_TRACK_GAIN") info.track_gain = value;
+				else if (id.ToUpper() == "REPLAYGAIN_TRACK_PEAK") info.track_peak = value;
+				else if (id.ToUpper() == "REPLAYGAIN_ALBUM_GAIN") info.album_gain = value;
+				else if (id.ToUpper() == "REPLAYGAIN_ALBUM_PEAK") info.album_peak = value;
+			}
+			else if (id == g_wszWMMCDI)
 			{
 				/* Use a heuristic to detect if this is a valid binary MCDI
 				 * field or the commonly used track offset string.
@@ -495,19 +536,11 @@ Error BoCA::TaggerWMA::ParseStreamInfo(const String &fileName, Track &track)
 
 					memcpy(data, pbValue, cbLength);
 
-					/* Check validity of MCDI data.
+					/* Check validity of MCDI data and assign.
 					 */
 					MCDI	 mcdi(data);
-					Bool	 valid = True;
 
-					for (Int i = 1; i < mcdi.GetNumberOfEntries(); i++)
-					{
-						if (mcdi.GetNthEntryOffset(i - 1) >= mcdi.GetNthEntryOffset(i)) valid = False;
-					}
-
-					/* Found a binary MCDI field.
-					 */
-					if (valid) info.mcdi.SetData(data);
+					if (mcdi.IsValid()) info.mcdi.SetData(data);
 				}
 				else
 				{
@@ -521,7 +554,7 @@ Error BoCA::TaggerWMA::ParseStreamInfo(const String &fileName, Track &track)
 					}
 				}
 			}
-			else if (String(name) == g_wszWMPicture && config->GetIntValue(ConfigID, "CoverArtReadFromTags", True))
+			else if (id == g_wszWMPicture && config->GetIntValue(ConfigID, "CoverArtReadFromTags", True))
 			{
 				WM_PICTURE	*picData = (WM_PICTURE *) pbValue;
 				Picture		 picture;
@@ -662,9 +695,7 @@ Error BoCA::TaggerWMA::UpdateStreamInfo(const String &fileName, const Track &tra
 
 	/* Open file and remove tags.
 	 */
-	String	 uncPath = String(fileName.StartsWith("\\\\") ? "" : uncPrefix).Append(fileName);
-
-	hr = metadataEditor2->OpenEx(uncPath, GENERIC_READ | GENERIC_WRITE, 0);
+	hr = metadataEditor2->OpenEx(toUNCPath(fileName), GENERIC_READ | GENERIC_WRITE, 0);
 
 	if (!FAILED(hr))
 	{
@@ -723,6 +754,12 @@ Error BoCA::TaggerWMA::UpdateStreamInfo(const String &fileName, const Track &tra
 
 				    nameStr == g_wszWMCopyright			||
 
+				    nameStr == "WM/Media"			||
+				    nameStr == "WM/CatalogNo"			||
+				    nameStr == "WM/Barcode"			||
+
+				    nameStr == g_wszWMLyrics			||
+
 				    nameStr == g_wszWMRadioStationName		||
 				    nameStr == g_wszWMRadioStationOwner		||
 
